"""
Implementing a basic command-line interface.
"""

import argparse
import logging
import sys

try:  # Python 3.8+
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

from platform import python_version

from .cli_utils import (cli_crawler, cli_discovery, examine,
                        file_processing_pipeline, load_blacklist,
                        load_input_dict, probe_homepage,
                        url_processing_pipeline, write_result)
from .settings import PARALLEL_CORES

# fix output encoding on some systems
try:
    # > Python 3.7
    if sys.stdout.encoding != 'UTF-8':
        sys.stdout.reconfigure(encoding='utf-8')
    if sys.stderr.encoding != 'UTF-8':
        sys.stderr.reconfigure(encoding='utf-8')
except AttributeError:
    import codecs
    if sys.stdout.encoding != 'UTF-8':
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'UTF-8':
        sys.stderr = codecs.getwriter('utf-8')(sys.stderr.buffer, 'strict')


def add_args(parser):
    "Add argument groups and arguments to parser."

    group1 = parser.add_argument_group('Input', 'URLs, files or directories to process')
    group1_ex = group1.add_mutually_exclusive_group()
    group2 = parser.add_argument_group('Output', 'Determines if and how files will be written')
    group3 = parser.add_argument_group('Navigation', 'Link discovery and web crawling')
    group3_ex = group3.add_mutually_exclusive_group()
    group4 = parser.add_argument_group('Extraction', 'Customization of text and metadata processing')
    group5 = parser.add_argument_group('Format', 'Selection of the output format')
    group5_ex = group5.add_mutually_exclusive_group()

    group1_ex.add_argument("-i", "--input-file",
                        help="name of input file for batch processing",
                        type=str)
    group1_ex.add_argument("--inputfile",
                        help=argparse.SUPPRESS,
                        type=str)   # will be deprecated
    group1_ex.add_argument("--input-dir",
                        help="read files from a specified directory (relative path)",
                        type=str)
    group1_ex.add_argument("--inputdir",
                        help=argparse.SUPPRESS,
                        type=str)   # will be deprecated
    group1_ex.add_argument("-u", "--URL",
                        help="custom URL download",
                        type=str)

    group1.add_argument('--parallel',
                        help="specify a number of cores/threads for downloads and/or processing",
                        type=int, default=PARALLEL_CORES)
    group1.add_argument('-b', '--blacklist',
                        help="file containing unwanted URLs to discard during processing",
                        type=str)

    group2.add_argument("--list",
                        help="display a list of URLs without downloading them",
                        action="store_true")
    group2.add_argument("-o", "--output-dir",
                        help="write results in a specified directory (relative path)",
                        type=str)
    group2.add_argument("--outputdir",
                        help=argparse.SUPPRESS,
                        type=str)   # will be deprecated
    group2.add_argument('--backup-dir',
                        help="preserve a copy of downloaded files in a backup directory",
                        type=str)
    group2.add_argument('--keep-dirs',
                        help="keep input directory structure and file names",
                        action="store_true")
    group2.add_argument('--hash-as-name',
                        help=argparse.SUPPRESS,
                        action="store_true")   # will be deprecated

    group3_ex.add_argument("--feed",
                        help="look for feeds and/or pass a feed URL as input",
                        nargs='?', const=True, default=False)
    group3_ex.add_argument("--sitemap",
                        help="look for sitemaps for the given website and/or enter a sitemap URL",
                        nargs='?', const=True, default=False)
    group3_ex.add_argument("--crawl",
                        help="crawl a fixed number of pages within a website starting from the given URL",
                        nargs='?', const=True, default=False)
    group3_ex.add_argument("--explore",
                        help="explore the given websites (combination of sitemap and crawl)",
                        nargs='?', const=True, default=False)
    group3_ex.add_argument("--probe",
                        help="probe for extractable content (works best with target language)",
                        nargs='?', const=True, default=False)
    group3.add_argument('--archived',
                        help='try to fetch URLs from the Internet Archive if downloads fail',
                        action="store_true")
    group3.add_argument('--url-filter',
                        help="only process/output URLs containing these patterns (space-separated strings)",
                        nargs='+', type=str)
    #group3.add_argument('--no-ssl',
    #                    help="Disable secure connections (to prevent SSLError)",
    #                    action="store_true")

    group4.add_argument("-f", "--fast",
                        help="fast (without fallback detection)",
                        action="store_true")
    group4.add_argument("--formatting",
                        help="include text formatting (bold, italic, etc.)",
                        action="store_true")
    group4.add_argument("--links",
                        help="include links along with their targets (experimental)",
                        action="store_true")
    group4.add_argument("--images",
                        help="include image sources in output (experimental)",
                        action="store_true")
    group4.add_argument("--nocomments",
                        help=argparse.SUPPRESS,
                        action="store_false")  # will be deprecated
    group4.add_argument("--notables",
                        help=argparse.SUPPRESS,
                        action="store_false")  # will be deprecated
    group4.add_argument("--no-comments",
                        help="don't output any comments",
                        action="store_false")  # false = no comments
    group4.add_argument("--no-tables",
                        help="don't output any table elements",
                        action="store_false")  # false = no tables
    group4.add_argument("--only-with-metadata",
                        help="only output those documents with title, URL and date (for formats supporting metadata)",
                        action="store_true")
    group4.add_argument("--with-metadata",
                        help=argparse.SUPPRESS,
                        action="store_true")   # will be deprecated
    group4.add_argument("--target-language",
                        help="select a target language (ISO 639-1 codes)",
                        type=str)
    group4.add_argument("--deduplicate",
                        help="filter out duplicate documents and sections",
                        action="store_true")
    group4.add_argument("--config-file",
                        help="override standard extraction parameters with a custom config file",
                        type=str)
    group4.add_argument("--precision",
                        help="favor extraction precision (less noise, possibly less text)",
                        action="store_true")
    group4.add_argument("--recall",
                        help="favor extraction recall (more text, possibly more noise)",
                        action="store_true")

    # https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_mutually_exclusive_group
    group5_ex.add_argument('-out', '--output-format',
                        help="determine output format",
                        choices=['txt', 'csv', 'html', 'json', 'markdown', 'xml', 'xmltei'],
                        default='txt')
    group5_ex.add_argument("--csv",
                        help="shorthand for CSV output",
                        action="store_true")
    group5_ex.add_argument("--html",
                        help="shorthand for HTML output",
                        action="store_true")
    group5_ex.add_argument("--json",
                        help="shorthand for JSON output",
                        action="store_true")
    group5_ex.add_argument("--markdown",
                        help="shorthand for MD output",
                        action="store_true")
    group5_ex.add_argument("--xml",
                        help="shorthand for XML output",
                        action="store_true")
    group5_ex.add_argument("--xmltei",
                        help="shorthand for XML TEI output",
                        action="store_true")
    group5.add_argument("--validate-tei",
                        help="validate XML TEI output",
                        action="store_true")

    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="increase logging verbosity (-v or -vv)",
                        )
    parser.add_argument(
        "--version",
        help="show version information and exit",
        action="version",
        version=f"Trafilatura {version('trafilatura')} - Python {python_version()}",
    )

    return parser


def parse_args(args):
    """Define parser for command-line arguments"""
    parser = argparse.ArgumentParser(description='Command-line interface for Trafilatura')
    parser = add_args(parser)
    # wrap in mapping to prevent invalid input
    return map_args(parser.parse_args())


def map_args(args):
    '''Map existing options to format and output choices.'''
    # formats
    for otype in ("csv", "html", "json", "markdown", "xml", "xmltei"):
        if getattr(args, otype):
            args.output_format = otype
            break
    # output configuration
    if args.nocomments is False:
        raise ValueError(
              "--nocomments is deprecated, use --no-comments instead",
              )
    if args.notables is False:
        raise ValueError(
              "--notables is deprecated, use --no-tables instead",
              )
    if args.inputfile:
        raise ValueError(
              "--inputfile is deprecated, use --input-file instead",
              )
    if args.inputdir:
        raise ValueError(
              "--inputdir is deprecated, use --input-dir instead",
              )
    if args.outputdir:
        raise ValueError(
              "--outputdir is deprecated, use --output-dir instead",
              )
    if args.hash_as_name:
        raise ValueError(
              "--hash-as-name is deprecated, hashes are used by default",
              )
    return args


def main():
    """ Run as a command-line utility. """
    args = parse_args(sys.argv[1:])
    process_args(args)


def process_args(args):
    """Perform the actual processing according to the arguments"""
    # init
    error_caught = False
    # verbosity
    if args.verbose == 1:
        logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
    elif args.verbose >= 2:
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    if args.blacklist:
        args.blacklist = load_blacklist(args.blacklist)

    # processing according to mutually exclusive options
    # read url list from input file
    if args.input_file and all([not args.crawl, not args.explore, not args.feed, not args.probe, not args.sitemap]):
        url_store = load_input_dict(args)
        error_caught = url_processing_pipeline(args, url_store)

    # fetch urls from a feed or a sitemap
    elif args.explore or args.feed or args.sitemap:
        cli_discovery(args)

    # activate crawler/spider
    elif args.crawl:
        cli_crawler(args)

    # probe and print only
    elif args.probe:
        probe_homepage(args)

    # read files from an input directory
    elif args.input_dir:
        file_processing_pipeline(args)

    # process input URL
    elif args.URL:
        url_store = load_input_dict(args)
        error_caught = url_processing_pipeline(args, url_store)  # process single url

    # read input on STDIN directly
    else:
        result = examine(sys.stdin.buffer.read(), args, url=args.URL)
        write_result(result, args)

    # change exit code if there are errors
    if error_caught is True:
        sys.exit(1)


if __name__ == '__main__':
    main()



# pylint:disable-msg=E0611
"""
X-Path expressions used to extract or filter the main text content,
and to extract metadata.
"""

from lxml.etree import XPath


### 1. CONTENT


BODY_XPATH = [XPath(x) for x in (
    '''.//*[(self::article or self::div or self::main or self::section)][
    @class="post" or @class="entry" or
    contains(@class, "post-text") or contains(@class, "post_text") or
    contains(@class, "post-body") or contains(@class, "post-entry") or contains(@class, "postentry") or
    contains(@class, "post-content") or contains(@class, "post_content") or
    contains(@class, "postcontent") or contains(@class, "postContent") or contains(@class, "post_inner_wrapper") or
    contains(@class, "article-text") or contains(@class, "articletext") or contains(@class, "articleText")
    or contains(@id, "entry-content") or
    contains(@class, "entry-content") or contains(@id, "article-content") or
    contains(@class, "article-content") or contains(@id, "article__content") or
    contains(@class, "article__content") or contains(@id, "article-body") or
    contains(@class, "article-body") or contains(@id, "article__body") or
    contains(@class, "article__body") or @itemprop="articleBody" or
    contains(translate(@id, "B", "b"), "articlebody") or contains(translate(@class, "B", "b"), "articleBody")
    or @id="articleContent" or contains(@class, "ArticleContent") or
    contains(@class, "page-content") or contains(@class, "text-content") or
    contains(@id, "body-text") or contains(@class, "body-text") or
    contains(@class, "article__container") or contains(@id, "art-content") or contains(@class, "art-content")][1]''',
    # (…)[1] = first occurrence
    '(.//article)[1]',
    """(.//*[(self::article or self::div or self::main or self::section)][
    contains(@class, 'post-bodycopy') or
    contains(@class, 'storycontent') or contains(@class, 'story-content') or
    @class='postarea' or @class='art-postcontent' or
    contains(@class, 'theme-content') or contains(@class, 'blog-content') or
    contains(@class, 'section-content') or contains(@class, 'single-content') or
    contains(@class, 'single-post') or
    contains(@class, 'main-column') or contains(@class, 'wpb_text_column') or
    starts-with(@id, 'primary') or starts-with(@class, 'article ') or @class="text" or
    @id="article" or @class="cell" or @id="story" or @class="story" or
    contains(@class, "story-body") or contains(@id, "story-body") or contains(@class, "field-body") or
    contains(translate(@class, "FULTEX","fultex"), "fulltext")
    or @role='article'])[1]""",
    '''(.//*[(self::article or self::div or self::main or self::section)][
    contains(@id, "content-main") or contains(@class, "content-main") or contains(@class, "content_main") or
    contains(@id, "content-body") or contains(@class, "content-body") or contains(@id, "contentBody")
    or contains(@class, "content__body") or contains(translate(@id, "CM","cm"), "main-content") or contains(translate(@class, "CM","cm"), "main-content")
    or contains(translate(@class, "CP","cp"), "page-content") or
    @id="content" or @class="content"])[1]''',
    '(.//*[(self::article or self::div or self::section)][starts-with(@class, "main") or starts-with(@id, "main") or starts-with(@role, "main")])[1]|(.//main)[1]',
)]
# starts-with(@id, "article") or
# or starts-with(@id, "story") or contains(@class, "story")
# starts-with(@class, "content ") or contains(@class, " content")
# '//div[contains(@class, "text") or contains(@class, "article-wrapper") or contains(@class, "content-wrapper")]',
# '//div[contains(@class, "article-wrapper") or contains(@class, "content-wrapper")]',
# |//*[(self::article or self::div or self::main or self::section)][contains(@class, "article") or contains(@class, "Article")]
# @id="content"or @class="content" or @class="Content"
# or starts-with(@class, 'post ')
# './/span[@class=""]', # instagram?


COMMENTS_XPATH = [XPath(x) for x in (
    """.//*[(self::div or self::list or self::section)][contains(@id, 'commentlist')
    or contains(@class, 'commentlist') or contains(@class, 'comment-page') or
    contains(@id, 'comment-list') or contains(@class, 'comments-list') or
    contains(@class, 'comments-content') or contains(@class, 'post-comments')]""",
    """.//*[(self::div or self::section or self::list)][starts-with(@id, 'comments')
    or starts-with(@class, 'comments') or starts-with(@class, 'Comments') or
    starts-with(@id, 'comment-') or starts-with(@class, 'comment-') or
    contains(@class, 'article-comments')]""",
    """.//*[(self::div or self::section or self::list)][starts-with(@id, 'comol') or
    starts-with(@id, 'disqus_thread') or starts-with(@id, 'dsq-comments')]""",
    ".//*[(self::div or self::section)][starts-with(@id, 'social') or contains(@class, 'comment')]",
)]
# or contains(@class, 'Comments')


REMOVE_COMMENTS_XPATH = [XPath(
    """.//*[(self::div or self::list or self::section)][
    starts-with(translate(@id, "C","c"), 'comment') or
    starts-with(translate(@class, "C","c"), 'comment') or
    contains(@class, 'article-comments') or contains(@class, 'post-comments')
    or starts-with(@id, 'comol') or starts-with(@id, 'disqus_thread')
    or starts-with(@id, 'dsq-comments')
    ]"""
)]
# or self::span
# or contains(@class, 'comment') or contains(@id, 'comment')


PAYWALL_DISCARD_XPATH = [XPath(
    '''.//*[(self::div or self::p)][
    contains(@id, "paywall") or contains(@id, "premium") or
    contains(@class, "paid-content") or contains(@class, "paidcontent") or
    contains(@class, "obfuscated") or contains(@class, "blurred") or
    contains(@class, "restricted") or contains(@class, "overlay")
    ]'''
)]


OVERALL_DISCARD_XPATH = [XPath(x) for x in (
    # navigation + footers, news outlets related posts, sharing, jp-post-flair jp-relatedposts
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
    contains(translate(@id, "F","f"), "footer") or contains(translate(@class, "F","f"), "footer")
    or contains(@id, "related") or contains(translate(@class, "R", "r"), "related") or
    contains(@id, "viral") or contains(@class, "viral") or
    starts-with(@id, "shar") or starts-with(@class, "shar") or
    contains(@class, "share-") or
    contains(translate(@id, "S", "s"), "share") or
    contains(@id, "social") or contains(@class, "social") or contains(@class, "sociable") or
    contains(@id, "syndication") or contains(@class, "syndication") or
    starts-with(@id, "jp-") or starts-with(@id, "dpsp-content") or
    contains(@class, "embedded") or contains(@class, "embed")
    or contains(@id, "newsletter") or contains(@class, "newsletter")
    or contains(@class, "subnav") or
    contains(@id, "cookie") or contains(@class, "cookie") or
    contains(@id, "tags") or contains(@class, "tags")  or contains(@class, "tag-list") or
    contains(@id, "sidebar") or contains(@class, "sidebar") or
    contains(@id, "banner") or contains(@class, "banner") or contains(@class, "bar") or
    contains(@class, "meta") or contains(@id, "menu") or contains(@class, "menu") or
    contains(translate(@id, "N", "n"), "nav") or contains(translate(@role, "N", "n"), "nav")
    or starts-with(@class, "nav") or contains(translate(@class, "N", "n"), "navigation") or
    contains(@class, "navbar") or contains(@class, "navbox") or starts-with(@class, "post-nav")
    or contains(@id, "breadcrumb") or contains(@class, "breadcrumb") or
    contains(@id, "bread-crumb") or contains(@class, "bread-crumb") or
    contains(@id, "author") or contains(@class, "author") or
    contains(@id, "button") or contains(@class, "button")
    or contains(translate(@class, "B", "b"), "byline")
    or contains(@class, "rating") or contains(@class, "widget") or
    contains(@class, "attachment") or contains(@class, "timestamp") or
    contains(@class, "user-info") or contains(@class, "user-profile") or
    contains(@class, "-ad-") or contains(@class, "-icon")
    or contains(@class, "article-infos") or
    contains(translate(@class, "I", "i"), "infoline")
    or contains(@data-component, "MostPopularStories")
    or contains(@class, "outbrain") or contains(@class, "taboola")
    or contains(@class, "criteo") or contains(@class, "options") or contains(@class, "expand")
    or contains(@class, "consent") or contains(@class, "modal-content")
    or contains(@class, "paid-content") or contains(@class, "paidcontent")
    or contains(@id, "premium-") or contains(@id, "paywall")
    or contains(@class, "obfuscated") or contains(@class, "blurred")
    or contains(@class, " ad ")
    or contains(@class, "next-") or contains(@class, "side-stories")
    or contains(@class, "related-stories") or contains(@class, "most-popular")
    or contains(@class, "mol-factbox") or starts-with(@class, "ZendeskForm")
    or contains(@class, "message-container") or contains(@id, "message_container")
    or contains(@class, "yin") or contains(@class, "zlylin") or
    contains(@class, "xg1") or contains(@id, "bmdh") or
    contains(@class, "slide") or contains(@class, "viewport") or
    @data-lp-replacement-content or @data-testid]''',

    # comment debris + hidden parts
    '''.//*[@class="comments-title" or contains(@class, "comments-title") or
    contains(@class, "nocomments") or starts-with(@id, "reply-") or starts-with(@class, "reply-") or
    contains(@class, "-reply-") or contains(@class, "message") or contains(@id, "reader-comments")
    or contains(@id, "akismet") or contains(@class, "akismet") or contains(@class, "suggest-links") or
    starts-with(@class, "hide-") or contains(@class, "-hide-") or contains(@class, "hide-print") or
    contains(@id, "hidden") or contains(@style, "hidden") or contains(@class, " hidden") or
    contains(@class, "noprint") or contains(@style, "display:none") or contains(@style, "display: none")
    or @aria-hidden="true" or contains(@class, "notloaded")]''',
)]
# conflicts:
# contains(@id, "header") or contains(@class, "header") or
# class contains "cats" (categories, also tags?)
# or contains(@class, "hidden ")  or contains(@class, "-hide")
# or contains(@class, "paywall")
# contains(@class, "content-info") or contains(@class, "content-title")
# contains(translate(@class, "N", "n"), "nav") or
# contains(@class, "panel") or
# or starts-with(@id, "comment-")


# the following conditions focus on extraction precision
TEASER_DISCARD_XPATH = [XPath(
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
        contains(translate(@id, "T", "t"), "teaser") or contains(translate(@class, "T", "t"), "teaser")
    ]'''
)]


PRECISION_DISCARD_XPATH = [XPath(x) for x in (
    './/header',
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
        contains(@id, "bottom") or contains(@class, "bottom") or
        contains(@id, "link") or contains(@class, "link")
        or contains(@style, "border")
    ]''',
)]
# or contains(@id, "-comments") or contains(@class, "-comments")


DISCARD_IMAGE_ELEMENTS = [XPath(
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
             contains(@id, "caption") or contains(@class, "caption")
            ]
    '''
)]


COMMENTS_DISCARD_XPATH = [XPath(x) for x in (
    './/*[(self::div or self::section)][starts-with(@id, "respond")]',
    './/cite|.//quote',
    '''.//*[@class="comments-title" or contains(@class, "comments-title") or
    contains(@class, "nocomments") or starts-with(@id, "reply-") or
    starts-with(@class, "reply-") or contains(@class, "-reply-") or contains(@class, "message")
    or contains(@class, "signin") or
    contains(@id, "akismet") or contains(@class, "akismet") or contains(@style, "display:none")]''',
)]



### 2. METADATA


# the order or depth of XPaths could be changed after exhaustive testing
AUTHOR_XPATHS = [XPath(x) for x in (
    # specific and almost specific
    '//*[(self::a or self::address or self::div or self::link or self::p or self::span or self::strong)][@rel="author" or @id="author" or @class="author" or @itemprop="author name" or rel="me" or contains(@class, "author-name") or contains(@class, "AuthorName") or contains(@class, "authorName") or contains(@class, "author name") or @data-testid="AuthorCard" or @data-testid="AuthorURL"]|//author',
    # almost generic and generic, last ones not common
    '//*[(self::a or self::div or self::h3 or self::h4 or self::p or self::span)][contains(@class, "author") or contains(@id, "author") or contains(@itemprop, "author") or @class="byline" or contains(@class, "channel-name") or contains(@id, "zuozhe") or contains(@class, "zuozhe") or contains(@id, "bianji") or contains(@class, "bianji") or contains(@id, "xiaobian") or contains(@class, "xiaobian") or contains(@class, "submitted-by") or contains(@class, "posted-by") or @class="username" or @class="byl" or @class="BBL" or contains(@class, "journalist-name")]',
     # last resort: any element
    '//*[contains(translate(@id, "A", "a"), "author") or contains(translate(@class, "A", "a"), "author") or contains(@class, "screenname") or contains(@data-component, "Byline") or contains(@itemprop, "author") or contains(@class, "writer") or contains(translate(@class, "B", "b"), "byline")]',
)]


AUTHOR_DISCARD_XPATHS = [XPath(x) for x in (
    """.//*[(self::a or self::div or self::section or self::span)][@id='comments' or @class='comments' or @class='title' or @class='date' or
    contains(@id, 'commentlist') or contains(@class, 'commentlist') or contains(@class, 'sidebar') or contains(@class, 'is-hidden') or contains(@class, 'quote')
    or contains(@id, 'comment-list') or contains(@class, 'comments-list') or contains(@class, 'embedly-instagram') or contains(@id, 'ProductReviews') or
    starts-with(@id, 'comments') or contains(@data-component, "Figure") or contains(@class, "article-share") or contains(@class, "article-support") or contains(@class, "print") or contains(@class, "category") or contains(@class, "meta-date") or contains(@class, "meta-reviewer")
    or starts-with(@class, 'comments') or starts-with(@class, 'Comments')
    ]""",
    '//time|//figure',
)]


CATEGORIES_XPATHS = [XPath(x) for x in (
    """//div[starts-with(@class, 'post-info') or starts-with(@class, 'postinfo') or
    starts-with(@class, 'post-meta') or starts-with(@class, 'postmeta') or
    starts-with(@class, 'meta') or starts-with(@class, 'entry-meta') or starts-with(@class, 'entry-info') or
    starts-with(@class, 'entry-utility') or starts-with(@id, 'postpath')]//a[@href]""",
    "//p[starts-with(@class, 'postmeta') or starts-with(@class, 'entry-categories') or @class='postinfo' or @id='filedunder']//a[@href]",
    "//footer[starts-with(@class, 'entry-meta') or starts-with(@class, 'entry-footer')]//a[@href]",
    '//*[(self::li or self::span)][@class="post-category" or @class="postcategory" or @class="entry-category" or contains(@class, "cat-links")]//a[@href]',
    '//header[@class="entry-header"]//a[@href]',
    '//div[@class="row" or @class="tags"]//a[@href]',
)]
# "//*[(self::div or self::p)][contains(@class, 'byline')]",


TAGS_XPATHS = [XPath(x) for x in (
    '//div[@class="tags"]//a[@href]',
    "//p[starts-with(@class, 'entry-tags')]//a[@href]",
    '''//div[@class="row" or @class="jp-relatedposts" or
    @class="entry-utility" or starts-with(@class, 'tag') or
    starts-with(@class, 'postmeta') or starts-with(@class, 'meta')]//a[@href]''',
    '//*[@class="entry-meta" or contains(@class, "topics") or contains(@class, "tags-links")]//a[@href]',
)]
# "related-topics"
# https://github.com/grangier/python-goose/blob/develop/goose/extractors/tags.py


TITLE_XPATHS = [XPath(x) for x in (
    '//*[(self::h1 or self::h2)][contains(@class, "post-title") or contains(@class, "entry-title") or contains(@class, "headline") or contains(@id, "headline") or contains(@itemprop, "headline") or contains(@class, "post__title") or contains(@class, "article-title")]',
    '//*[@class="entry-title" or @class="post-title"]',
    '//*[(self::h1 or self::h2 or self::h3)][contains(@class, "title") or contains(@id, "title")]',
)]
# json-ld headline
# '//header/h1',



# pylint:disable-msg=E0611,I1101
"""
All functions related to XML generation, processing and validation.
"""

import csv
import logging
import lzma

from html import unescape
from io import StringIO
from json import dumps as json_dumps
from pathlib import Path
from pickle import load as load_pickle

try:  # Python 3.8+
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

from lxml.etree import (Element, RelaxNG, SubElement, XMLParser, fromstring,
                        tostring)

from .utils import sanitize, sanitize_tree, text_chars_test


LOGGER = logging.getLogger(__name__)
PKG_VERSION = version("trafilatura")

# validation
TEI_SCHEMA = str(Path(__file__).parent / 'data/tei-schema-pickle.lzma')
TEI_VALID_TAGS = {'ab', 'body', 'cell', 'code', 'del', 'div', 'graphic', 'head', 'hi', \
                  'item', 'lb', 'list', 'p', 'quote', 'ref', 'row', 'table'}
TEI_VALID_ATTRS = {'rend', 'rendition', 'role', 'target', 'type'}
TEI_RELAXNG = None  # to be downloaded later if necessary
TEI_REMOVE_TAIL = {"ab", "p"}
TEI_DIV_SIBLINGS = {"p", "list", "table", "quote", "ab"}

CONTROL_PARSER = XMLParser(remove_blank_text=True)

NEWLINE_ELEMS = {
    **{tag: '\n' for tag in ['code', 'graphic', 'head', 'lb', 'list', 'p', 'quote', 'row', 'table']}
}
SPECIAL_FORMATTING = {'del', 'head', 'hi', 'ref'}
WITH_ATTRIBUTES = {'cell', 'row', 'del', 'graphic', 'head', 'hi', 'item', 'list', 'ref'}

NESTING_WHITELIST = {"cell", "figure", "item", "note", "quote"}

META_ATTRIBUTES = [
    'sitename', 'title', 'author', 'date', 'url', 'hostname',
    'description', 'categories', 'tags', 'license', 'id',
    'fingerprint', 'language'
]

HI_FORMATTING = {'#b': '**', '#i': '*', '#u': '__', '#t': '`'}


def build_json_output(docmeta, with_metadata=True):
    '''Build JSON output based on extracted information'''
    if with_metadata:
        outputdict = {slot: getattr(docmeta, slot, None) for slot in docmeta.__slots__}
        outputdict.update({
            'source': outputdict.pop('url'),
            'source-hostname': outputdict.pop('sitename'),
            'excerpt': outputdict.pop('description'),
            'categories': ';'.join(outputdict.pop('categories')),
            'tags': ';'.join(outputdict.pop('tags')),
            'text': xmltotxt(outputdict.pop('body'), include_formatting=False),
        })
        commentsbody = outputdict.pop('commentsbody')
    else:
        outputdict = {'text': xmltotxt(docmeta.body, include_formatting=False)}
        commentsbody = docmeta.commentsbody

    if commentsbody is not None:
        outputdict['comments'] = xmltotxt(commentsbody, include_formatting=False)

    return json_dumps(outputdict, ensure_ascii=False)


def clean_attributes(tree):
    '''Remove unnecessary attributes.'''
    for elem in tree.iter('*'):
        if elem.tag not in WITH_ATTRIBUTES:
            elem.attrib.clear()
    return tree


def remove_empty_elements(tree):
    '''Remove text elements without text.'''
    for element in tree.iter('*'):  # 'head', 'hi', 'item', 'p'
        if len(element) == 0 and text_chars_test(element.text) is False and text_chars_test(element.tail) is False:
            parent = element.getparent()
            # not root element or element which is naturally empty
            # do not remove elements inside <code> to preserve formatting
            if parent is not None and element.tag != "graphic" and parent.tag != 'code':
                element.getparent().remove(element)
    return tree


def strip_double_tags(tree):
    "Prevent nested tags among a fixed list of tags."
    for elem in reversed(tree.xpath(".//head | .//code | .//p")):
        for subelem in elem.iterdescendants("code", "head", "p"):
            if subelem.getparent().tag in NESTING_WHITELIST:
                continue
            if subelem.tag == elem.tag:
                merge_with_parent(subelem)
    return tree


def build_xml_output(docmeta):
    '''Build XML output tree based on extracted information'''
    output = Element('doc')
    output = add_xml_meta(output, docmeta)
    docmeta.body.tag = 'main'
    # clean XML tree
    output.append(clean_attributes(docmeta.body))
    if docmeta.commentsbody is not None:
        docmeta.commentsbody.tag = 'comments'
        output.append(clean_attributes(docmeta.commentsbody))
# XML invalid characters
# https://chase-seibert.github.io/blog/2011/05/20/stripping-control-characters-in-python.html
    return output


def control_xml_output(document, options):
    '''Make sure the XML output is conform and valid if required'''
    strip_double_tags(document.body)
    remove_empty_elements(document.body)

    func = build_xml_output if options.format == "xml" else build_tei_output
    output_tree = func(document)

    output_tree = sanitize_tree(output_tree)
    # necessary for cleaning
    control_string = tostring(output_tree, encoding='unicode')
    output_tree = fromstring(control_string, CONTROL_PARSER)

    # validate
    if options.format == 'xmltei' and options.tei_validation:
        result = validate_tei(output_tree)
        LOGGER.debug('TEI validation result: %s %s', result, options.source)

    return tostring(output_tree, pretty_print=True, encoding='unicode').strip()


def add_xml_meta(output, docmeta):
    '''Add extracted metadata to the XML output tree'''
    for attribute in META_ATTRIBUTES:
        value = getattr(docmeta, attribute, None)
        if value:
            output.set(attribute, value if isinstance(value, str) else ';'.join(value))
    return output


def build_tei_output(docmeta):
    '''Build TEI-XML output tree based on extracted information'''
    # build TEI tree
    output = write_teitree(docmeta)
    # filter output (strip unwanted elements), just in case
    # check and repair
    output = check_tei(output, docmeta.url)
    return output


def check_tei(xmldoc, url):
    '''Check if the resulting XML file is conform and scrub remaining tags'''
    # convert head tags
    for elem in xmldoc.iter('head'):
        elem.tag = 'ab'
        elem.set('type', 'header')
        parent = elem.getparent()
        if len(elem) > 0:
            new_elem = _tei_handle_complex_head(elem)
            parent.replace(elem, new_elem)
            elem = new_elem
        if parent.tag == "p":
            _move_element_one_level_up(elem)
    # convert <lb/> when child of <div> to <p>
    for elem in xmldoc.findall(".//text/body//div/lb"):
        if elem.tail and elem.tail.strip():
            elem.tag, elem.text, elem.tail = 'p', elem.tail, None
    # look for elements that are not valid
    for elem in xmldoc.findall('.//text/body//*'):
        if elem.tag in TEI_REMOVE_TAIL and elem.tail and elem.tail.strip():
            _handle_unwanted_tails(elem)
        # check elements
        if elem.tag not in TEI_VALID_TAGS:
            # disable warnings for chosen categories
            # if element.tag not in ('div', 'span'):
            LOGGER.warning('not a TEI element, removing: %s %s', elem.tag, url)
            merge_with_parent(elem)
            continue
        if elem.tag == "div":
            _handle_text_content_of_div_nodes(elem)
            _wrap_unwanted_siblings_of_div(elem)
        # check attributes
        for attribute in elem.attrib:
            if attribute not in TEI_VALID_ATTRS:
                LOGGER.warning('not a valid TEI attribute, removing: %s in %s %s', attribute, elem.tag, url)
                elem.attrib.pop(attribute)
    return xmldoc


def validate_tei(xmldoc):  # , filename=""
    '''Check if an XML document is conform to the guidelines of the Text Encoding Initiative'''
    global TEI_RELAXNG
    if TEI_RELAXNG is None:
        # load validator
        with lzma.open(TEI_SCHEMA, 'rb') as schemafile:
            schema_data = load_pickle(schemafile)
        TEI_RELAXNG = RelaxNG(fromstring(schema_data))
    result = TEI_RELAXNG.validate(xmldoc)
    if result is False:
        LOGGER.warning('not a valid TEI document: %s', TEI_RELAXNG.error_log.last_error)
    return result


def replace_element_text(element, include_formatting):
    "Determine element text based on just the text of the element. One must deal with the tail separately."
    elem_text = element.text or ""
    # handle formatting: convert to markdown
    children = element.getchildren()
    if include_formatting and element.text:
        if element.tag == "head":
            try:
                number = int(element.get("rend")[1])
            except (TypeError, ValueError):
                number = 2
            elem_text = f'{"#" * number} {elem_text}'
        elif element.tag == "del":
            elem_text = f"~~{elem_text}~~"
        elif element.tag == "hi":
            rend = element.get("rend")
            if rend in HI_FORMATTING:
                elem_text = f"{HI_FORMATTING[rend]}{elem_text}{HI_FORMATTING[rend]}"
        elif element.tag == "code":
            if "\n" in element.text:
                elem_text = f"```\n{elem_text}\n```"
            else:
                elem_text = f"`{elem_text}`"
    # handle links
    if element.tag == "ref":
        if elem_text:
            link_text = f"[{elem_text}]"
            target = element.get("target")
            if target:
                elem_text = f"{link_text}({target})"
            else:
                LOGGER.warning("missing link attribute: %s %s'", elem_text, element.attrib)
                elem_text = link_text
        else:
            LOGGER.warning("empty link: %s %s", elem_text, element.attrib)
    # cells
    if element.tag == "cell" and elem_text and children and children[0].tag == 'p':
        elem_text = f"{elem_text} "
    # lists
    elif element.tag == "item" and elem_text:
        elem_text = f"- {elem_text}\n"
    return elem_text


def merge_with_parent(element, include_formatting=False):
    '''Merge element with its parent and convert formatting to markdown.'''
    parent = element.getparent()
    if parent is None:
        return

    full_text = replace_element_text(element, include_formatting)
    if element.tail is not None:
        full_text += element.tail

    previous = element.getprevious()
    if previous is not None:
        # There is a previous node, append text to its tail
        previous.tail = f'{previous.tail} {full_text}' if previous.tail else full_text
    elif parent.text is not None:
        parent.text = f'{parent.text} {full_text}'
    else:
        parent.text = full_text
    parent.remove(element)


def process_element(element, returnlist, include_formatting):
    # Process children recursively
    if element.text is not None:
        # this is the text that comes before the first child
        textelement = replace_element_text(element, include_formatting)
        returnlist.append(textelement)

    for child in element:
        process_element(child, returnlist, include_formatting)

    if element.text is None and element.tail is None:
        if element.tag == 'graphic':
            # add source, default to ''
            text = f'{element.get("title", "")} {element.get("alt", "")}'
            returnlist.extend(['![', text.strip(), ']', '(', element.get('src', ''), ')'])
        # newlines for textless elements
        if element.tag in NEWLINE_ELEMS:
            # add line after table head
            if element.tag == "row":
                max_span = int(element.get("span", 1))
                cell_count = len(element.xpath(".//cell"))
                # row ended so draw extra empty cells to match max_span
                returnlist.append("|" * (max_span - cell_count) + "\n")
                # if this is a head row, draw the separator below
                is_head = bool(element.xpath("./cell[@role='head']"))
                if is_head:
                    returnlist.append("\n" + "---|" * max_span + "\n")
            else:
                returnlist.append('\n')
        if element.tag != 'cell':
            # cells still need to append vertical bars
            # but nothing more to do with other textless elements
            return

    # Process text

    # Common elements (Now processes end-tag logic correctly)
    within_cell = element.xpath("ancestor::cell")
    if element.tag == 'p' and include_formatting:
        if not within_cell:
            returnlist.append('\n\u2424\n')
    elif element.tag in NEWLINE_ELEMS:
        if not within_cell:
            returnlist.extend([NEWLINE_ELEMS[element.tag], '\n'])
    elif element.tag == 'cell':
        returnlist.extend(" | ")
    elif element.tag == 'comments':
        returnlist.append('\n\n')
    else:
        if element.tag not in SPECIAL_FORMATTING:
            LOGGER.debug('unprocessed element in output: %s', element.tag)
            returnlist.extend([' '])

    # this is text that comes after the closing tag, so it should be after any NEWLINE_ELEMS
    if element.tail is not None:
        returnlist.append(element.tail)


def xmltotxt(xmloutput, include_formatting):
    '''Convert to plain text format and optionally preserve formatting as markdown.'''
    returnlist = []

    process_element(xmloutput, returnlist, include_formatting)

    return unescape(sanitize(''.join(returnlist)))


def xmltocsv(document, include_formatting, *, delim="\t", null="null"):
    "Convert the internal XML document representation to a CSV string."
    # preprocessing
    posttext = xmltotxt(document.body, include_formatting)
    if document.commentsbody is not None:
        commentstext = xmltotxt(document.commentsbody, include_formatting)
    else:
        commentstext = ""

    # output config
    output = StringIO()
    outputwriter = csv.writer(output, delimiter=delim, quoting=csv.QUOTE_MINIMAL)

    # organize fields
    data = (document.url,
            document.id,
            document.fingerprint,
            document.hostname,
            document.title,
            document.image,
            document.date,
            posttext,
            commentstext,
            document.license,
            document.pagetype)

    outputwriter.writerow([d if d else null for d in data])
    return output.getvalue()


def write_teitree(docmeta):
    '''Bundle the extracted post and comments into a TEI tree'''
    teidoc = Element('TEI', xmlns='http://www.tei-c.org/ns/1.0')
    header = write_fullheader(teidoc, docmeta)
    textelem = SubElement(teidoc, 'text')
    textbody = SubElement(textelem, 'body')
    # post
    postbody = clean_attributes(docmeta.body)
    postbody.tag = 'div'
    postbody.set('type', 'entry') # rendition='#pst'
    textbody.append(postbody)
    # comments
    if docmeta.commentsbody is not None:
        commentsbody = clean_attributes(docmeta.commentsbody)
        commentsbody.tag = 'div'
        commentsbody.set('type', 'comments') # rendition='#cmt'
        textbody.append(commentsbody)
    return teidoc


def _define_publisher_string(docmeta):
    '''Construct a publisher string to include in TEI header'''
    if docmeta.hostname and docmeta.sitename:
        publisherstring = f'{docmeta.sitename.strip()} ({docmeta.hostname})'
    elif docmeta.hostname:
        publisherstring = docmeta.hostname
    elif docmeta.sitename:
        publisherstring = docmeta.sitename
    else:
        LOGGER.warning('no publisher for URL %s', docmeta.url)
        publisherstring = 'N/A'
    return publisherstring


def write_fullheader(teidoc, docmeta):
    '''Write TEI header based on gathered metadata'''
    # todo: add language info
    header = SubElement(teidoc, 'teiHeader')
    filedesc = SubElement(header, 'fileDesc')
    bib_titlestmt = SubElement(filedesc, 'titleStmt')
    bib_titlemain = SubElement(bib_titlestmt, 'title', type='main')
    bib_titlemain.text = docmeta.title
    if docmeta.author:
        bib_author = SubElement(bib_titlestmt, 'author')
        bib_author.text = docmeta.author
    publicationstmt_a = SubElement(filedesc, 'publicationStmt')
    publisher_string = _define_publisher_string(docmeta)
    # license, if applicable
    if docmeta.license:
        publicationstmt_publisher = SubElement(publicationstmt_a, 'publisher')
        publicationstmt_publisher.text = publisher_string
        availability = SubElement(publicationstmt_a, 'availability')
        avail_p = SubElement(availability, 'p')
        avail_p.text = docmeta.license
    # insert an empty paragraph for conformity
    else:
        publicationstmt_p = SubElement(publicationstmt_a, 'p')
    notesstmt = SubElement(filedesc, 'notesStmt')
    if docmeta.id:
        idno = SubElement(notesstmt, 'note', type='id')
        idno.text = docmeta.id
    fingerprint = SubElement(notesstmt, 'note', type='fingerprint')
    fingerprint.text = docmeta.fingerprint
    sourcedesc = SubElement(filedesc, 'sourceDesc')
    source_bibl = SubElement(sourcedesc, 'bibl')
    # determination of sigle string
    if docmeta.sitename and docmeta.date:
        sigle = docmeta.sitename + ', ' + docmeta.date
    elif not docmeta.sitename and docmeta.date:
        sigle = docmeta.date
    elif docmeta.sitename:
        sigle = docmeta.sitename
    else:
        LOGGER.warning('no sigle for URL %s', docmeta.url)
        sigle = ''
    if docmeta.title:
        source_bibl.text = docmeta.title + '. ' + sigle
    else:
        source_bibl.text = '. ' + sigle
    source_sigle = SubElement(sourcedesc, 'bibl', type='sigle')
    source_sigle.text = sigle
    biblfull = SubElement(sourcedesc, 'biblFull')
    bib_titlestmt = SubElement(biblfull, 'titleStmt')
    bib_titlemain = SubElement(bib_titlestmt, 'title', type='main')
    bib_titlemain.text = docmeta.title
    if docmeta.author:
        bib_author = SubElement(bib_titlestmt, 'author')
        bib_author.text = docmeta.author
    publicationstmt = SubElement(biblfull, 'publicationStmt')
    publication_publisher = SubElement(publicationstmt, 'publisher')
    publication_publisher.text = publisher_string
    if docmeta.url:
        publication_url = SubElement(publicationstmt, 'ptr', type='URL', target=docmeta.url)
    publication_date = SubElement(publicationstmt, 'date')
    publication_date.text = docmeta.date
    profiledesc = SubElement(header, 'profileDesc')
    abstract = SubElement(profiledesc, 'abstract')
    abstract_p = SubElement(abstract, 'p')
    abstract_p.text = docmeta.description
    if len(docmeta.categories) > 0 or len(docmeta.tags) > 0:
        textclass = SubElement(profiledesc, 'textClass')
        keywords = SubElement(textclass, 'keywords')
        if len(docmeta.categories) > 0:
            cat_list = SubElement(keywords, 'term', type='categories')
            cat_list.text = ','.join(docmeta.categories)
        if len(docmeta.tags) > 0:
            tags_list = SubElement(keywords, 'term', type='tags')
            tags_list.text = ','.join(docmeta.tags)
    creation = SubElement(profiledesc, 'creation')
    date = SubElement(creation, 'date', type="download")
    date.text = docmeta.filedate
    encodingdesc = SubElement(header, 'encodingDesc')
    appinfo = SubElement(encodingdesc, 'appInfo')
    application = SubElement(appinfo, 'application', version=PKG_VERSION, ident='Trafilatura')
    label = SubElement(application, 'label')
    label.text = 'Trafilatura'
    pointer = SubElement(application, 'ptr', target='https://github.com/adbar/trafilatura')
    return header


def _handle_text_content_of_div_nodes(element):
    if element.text and element.text.strip():
        if element.getchildren() and element[0].tag == "p":
            element[0].text = f'{element.text} {element[0].text or ""}'.strip()
        else:
            new_child = Element("p")
            new_child.text = element.text
            element.insert(0, new_child)
        element.text = None

    if element.tail and element.tail.strip():
        if element.getchildren() and element[-1].tag == "p":
            element[-1].text = f'{element[-1].text or ""} {element.tail}'.strip()
        else:
            new_child = Element("p")
            new_child.text = element.tail
            element.append(new_child)
        element.tail = None


def _handle_unwanted_tails(element):
    "Handle tail on p and ab elements"
    if element.tag == "p":
        element.text = element.text + " " + element.tail.strip() if element.text else element.tail
    else:
        new_sibling = Element('p')
        new_sibling.text = element.tail.strip()
        parent = element.getparent()
        parent.insert(parent.index(element) + 1 , new_sibling)
    element.tail = None


def _tei_handle_complex_head(element):
    new_element = Element('ab', attrib=element.attrib)
    new_element.text = element.text.strip() if element.text is not None else None
    for child in element.iterchildren():
        if child.tag == 'p':
            if len(new_element) > 0 or new_element.text:
                # add <lb> if <ab> has no children or last tail contains text
                if len(new_element) == 0 or new_element[-1].tail:
                    SubElement(new_element, 'lb')
                new_element[-1].tail = child.text
            else:
                new_element.text = child.text
        else:
            new_element.append(child)
    if element.tail and element.tail.strip():
        new_element.tail = element.tail.strip()
    return new_element


def _wrap_unwanted_siblings_of_div(div_element):
    new_sibling = Element("div")
    new_sibling_index = None
    parent = div_element.getparent()
    # check siblings after target element
    for sibling in div_element.itersiblings():
        if sibling.tag == "div":
            break
        if sibling.tag in TEI_DIV_SIBLINGS:
            new_sibling_index = new_sibling_index or parent.index(sibling)
            new_sibling.append(sibling)
        # some elements (e.g. <lb/>) can appear next to div, but
        # order of elements should be kept, thus add and reset new_sibling
        else:
            if new_sibling_index and len(new_sibling) != 0:
                parent.insert(new_sibling_index, new_sibling)
                new_sibling = Element("div")
                new_sibling_index = None
    if new_sibling_index and len(new_sibling) != 0:
        parent.insert(new_sibling_index, new_sibling)


def _move_element_one_level_up(element):
    """
    Fix TEI compatibility issues by moving certain p-elems up in the XML tree.
    There is always a n+2 nesting for p-elements with the minimal structure ./TEI/text/body/p
    """
    parent = element.getparent()
    grand_parent = parent.getparent()

    new_elem = Element("p")
    new_elem.extend(sibling for sibling in element.itersiblings())

    grand_parent.insert(grand_parent.index(parent) + 1, element)

    if element.tail and element.tail.strip():
        new_elem.text = element.tail.strip()
        element.tail = None

    if parent.tail and parent.tail.strip():
        new_elem.tail = parent.tail.strip()
        parent.tail = None

    if len(new_elem) != 0 or new_elem.text or new_elem.tail:
        grand_parent.insert(grand_parent.index(element) + 1, new_elem)

    if len(parent) == 0 and parent.text is None:
        grand_parent.remove(parent)



import html
import json
import logging
import os
import re
from urllib.parse import unquote

from lxml import etree as ET
from py_asciimath.translator.translator import ASCIIMath2Tex
from resiliparse.parse.html import traverse_dom
from haruka_parser.utils import fast_html2text

logging.getLogger().setLevel(logging.ERROR)

color_regex = re.compile(r"\\textcolor\[.*?\]\{.*?\}")

asciimath2tex = ASCIIMath2Tex(log=False)

PARAGRAPH_TAGS = frozenset(
    {
        "body",
        "blockquote",
        "caption",
        "center",
        "col",
        "colgroup",
        "dd",
        "div",
        "dl",
        "dt",
        "fieldset",
        "form",
        "legend",
        "optgroup",
        "option",
        "p",
        "pre",
        "table",
        "td",
        "textarea",
        "tfoot",
        "th",
        "thead",
        "tr",
        "ul",
        "li",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
    }
)

latex_math_path = os.path.join(os.path.dirname(__file__), "dictionary/latex_words.txt")
with open(latex_math_path, "r") as f:
    latex_math_commands = [line.replace("\n", "").strip() for line in f]
    latex_math_commands = [line for line in latex_math_commands if line]

latex_image_class_names = [
    "latexcenter",
    "latex",
    "tex",
    "latexdisplay",
    "latexblock",
    "latexblockcenter",
]


latex_math_commands = [re.escape(term) for term in latex_math_commands]
latex_math_commands = [x + "(?![a-zA-Z])" for x in latex_math_commands]
latex_regex = re.compile("|".join(latex_math_commands))


def extract_asciimath(s):
    parsed = asciimath2tex.translate(s)
    return parsed


cur_file = os.path.abspath(__file__)
xsl_path = os.path.join(os.path.dirname(cur_file), "mmltex/mmltex.xsl")

xslt = ET.parse(xsl_path)
transform = ET.XSLT(xslt)


def mml_to_latex(mml_code):
    # Remove any attibutes from the math tag
    # Attention: replace into [itex] in the begining
    mml_code = mml_code.replace("[itex]", "<math>").replace("[/itex]", "</math>")
    mml_code = re.sub(r"(<math.*?>)", r"\1", mml_code)
    mml_ns = mml_code.replace(
        "<math>", '<math xmlns="http://www.w3.org/1998/Math/MathML">'
    )  # Required.
    mml_dom = ET.fromstring(mml_ns)
    mmldom = transform(mml_dom)
    latex_code = str(mmldom)
    return latex_code


def wrap_math(s, display=False):
    s = re.sub(r"\s+", " ", s)
    s = color_regex.sub("", s)
    s = s.strip()
    s = re.sub(r"^\$+", "", s)
    s = re.sub(r"\$+$", "", s)
    s = re.sub(r"^\\\[", "", s)
    s = re.sub(r"\\\]$", "", s)
    s = re.sub(r"^\\\(", "", s)
    s = re.sub(r"\\\)$", "", s)
    s = s.replace("\n", " ")
    s = s.strip()
    if len(s) == 0:
        return s
    # Don't wrap if it's already in \align
    # if "align" in s:
    #     return s
    if display:
        return "[extract_tex]" + s + "[/extract_tex]"
    return "[extract_itex]" + s + "[/extract_itex]"


def get_math_config(html):
    has_mathjax = re.search(r"mathjax", html.lower())
    has_katex = re.search(r"katex", html.lower())
    if not has_mathjax and not has_katex:
        return None
    # Get LaTeX config for MathJax
    regex = r"tex2jax: {[^}]*}"
    latex_config = {
        "inlineMath": [
            ["$", "$"],
            ["\\[", "\\]"],
            ["[itex]", "[/itex]"],
            ["[math]", "[/math]"],
            ["[latex]", "[/latex]"],
            ["[texx]", "[/texx]"],
        ],
        "displayMath": [["\\(", "\\)"], ["$$", "$$"], ["[tex]", "[/tex]"]],
        "skipTags": ["script", "noscript", "style", "textarea", "pre", "code"],
        "ignoreClass": "tex2jax_ignore",
    }
    try:
        match = re.search(regex, html)
        if match:
            config = match.group(0)
            # Make it a valid json object by adding quotes around the keys
            config = re.sub(r"(\w+):", r'"\1":', config)
            config = "{" + config + "}"
            # config = re.sub(r"\\", r"\\\\", config)
            config = re.sub(r"'", r'"', config)
            config = re.sub(r",\s*}", "}", config)
            extracted_latex_config = json.loads(config)["tex2jax"]
            # latex_config.update(extracted_latex_config)
            # Update this in a smart way: if the key is already there, append the values
            # if the key is not there, add it

            for key in extracted_latex_config:
                if key in latex_config and key != "ignoreClass":
                    latex_config[key] += extracted_latex_config[key]
                else:
                    latex_config[key] = extracted_latex_config[key]
    except Exception as e:
        pass

    # Get LaTeX config for KaTeX
    """      delimiters: [
          {left: '$$', right: '$$', display: true}
      ],
    """
    regex = r"delimiters: \[[^\]]*\]"
    try:
        match = re.search(regex, html)
        if match:
            config = match.group(0)
            # Make it a valid json object by adding quotes around the keys
            config = re.sub(r"(\w+):", r'"\1":', config)
            # The match is a list without the [] around it. Wrap with {"delimiters": ...}
            config = "{" + config + "}"
            config = re.sub(r"'", r'"', config)
            config = re.sub(r",\s*}", "}", config)
            extracted_latex_config = json.loads(config)["delimiters"]
            for delimiter in extracted_latex_config:
                if delimiter["display"]:
                    latex_config["displayMath"].append(
                        [delimiter["left"], delimiter["right"]]
                    )
                else:
                    latex_config["inlineMath"].append(
                        [delimiter["left"], delimiter["right"]]
                    )
    except Exception as e:
        pass

    # Get AsciiMath config
    regex = r"asciimath2jax: {[^}]*}"
    asciimath_config = {
        "delimiters": [["`", "`"]],
        "skipTags": ["script", "noscript", "style", "textarea", "pre", "code"],
        "ignoreClass": "asciimath2jax_ignore",
    }
    try:
        match = re.search(regex, html)
        if match:
            config = match.group(0)
            # Make it a valid json object by adding quotes around the keys
            config = re.sub(r"(\w+):", r'"\1":', config)
            config = "{" + config + "}"
            # config = re.sub(r"\\", r"\\\\", config)
            config = re.sub(r"'", r'"', config)
            config = re.sub(r",\s*}", "}", config)
            extracted_asciimath_config = json.loads(config)["asciimath2jax"]
            asciimath_config.update(extracted_asciimath_config)
    except Exception as e:
        pass
    return {"latex": latex_config, "asciimath": asciimath_config}


def html_unescape(s):
    return html.unescape(html.unescape(s))

def replace_bs_and_br(text):

    text = re.sub(r"\[HARUKA_PARSER_BS\]", " ", text)
    text = re.sub(r"\[HARUKA_PARSER_CHANGE_LINE\]", "\n", text)
    return text

def replace_math_tags_with_dollar_signs(text):
    # Replace each of these in the proper way
    # itex -> $...$
    # tex -> $$...$$
    # asciimath -> ...

    # Instead of this, simply replace extract_itex with $ and extract_tex with $$.
    text = re.sub(r"\[extract_itex\]", "$", text)
    text = re.sub(r"\[/extract_itex\]", "$", text)
    text = re.sub(r"\[extract_tex\]", "$$", text)
    text = re.sub(r"\[/extract_tex\]", "$$", text)

    return text


def update_text_with_delimiters(text, delimiters, replacement_manager, info):
    def replace_itex(match):
        wrapped = wrap_math(match.group(1))
        tagged = replacement_manager.add_replacement(wrapped, tag="math")
        return tagged

    def replace_tex(match):
        wrapped = wrap_math(match.group(1), display=True)
        tagged = replacement_manager.add_replacement(wrapped, tag="math")
        return tagged

    def replace_asciimath(match):
        wrapped = match.group(1)
        tagged = replacement_manager.add_replacement(wrapped, tag="math")
        return tagged

    for delimiter, type in delimiters:
        start_delimiter = re.escape(delimiter[0])
        end_delimiter = re.escape(delimiter[1])
        regex = f"{start_delimiter}(.*?){end_delimiter}"
        if type == "INLINE_LATEX":
            # Simply replace the delimiters with [itex] and [/itex]
            updated_text = re.sub(regex, replace_itex, text, flags=re.DOTALL)
            if updated_text != text:
                info["found_math"] = True
                info["mathjax_inline_tex"] += 1
            text = updated_text
        elif type == "DISPLAY_LATEX":
            updated_text = re.sub(regex, replace_tex, text, flags=re.DOTALL)
            if updated_text != text:
                info["found_math"] = True
                info["mathjax_display_tex"] += 1
            text = updated_text
        elif type == "ASCIIMATH":
            updated_text = re.sub(regex, replace_asciimath, text, flags=re.DOTALL)
            if updated_text != text:
                info["found_math"] = True
                info["mathjax_asciimath"] += 1
            text = updated_text

    return text

def update_math_parent_with_span(tree, math_node, math_text, replacement_manager, info, display=False):

    current_node = math_node
    parent_node = math_node.parent
    wrapped_math = wrap_math(math_text, display)
    while parent_node.tag in ["mathml", "mjx-container"]:
        current_node = parent_node
        parent_node = parent_node.parent
    new_span = tree.create_element("span")
    new_span.html = replacement_manager.add_replacement(
        wrapped_math, tag="math"
    )
    parent_node.replace_child(new_span, current_node)
    if len(wrapped_math.strip()) > 0:
        info["found_math"] = True

    return parent_node


def extract_delimited_math(text, mathjax_config, info, replacement_manager):
    """This operates on plain text and extracts LaTeX and AsciiMath"""
    # import pdb; pdb.set_trace()
    if mathjax_config is None:
        return text
    delimiters = []
    for delimiter in mathjax_config["latex"]["inlineMath"]:
        delimiters.append((delimiter, "INLINE_LATEX"))
    for delimiter in mathjax_config["latex"]["displayMath"]:
        delimiters.append((delimiter, "DISPLAY_LATEX"))
    for delimiter in mathjax_config["asciimath"]["delimiters"]:
        delimiters.append((delimiter, "ASCIIMATH"))

    delimiters = sorted(delimiters, key=lambda x: len(x[0][0]), reverse=True)
    text = update_text_with_delimiters(text, delimiters, replacement_manager, info)
    return text


def extract_math(tree, replacement_manager, info):
    """Webpages often contain LaTeX or AsciiMath equations that are
    hidden within the HTML. This function extracts the LaTeX and
    AsciiMath equations from the HTML.
    """

    # Find and tag any \align environments
    def start_callback(element):
        regex = r"\\begin{align}(.*?)\\end{align}"
        if element.node.type == 3:
            text = element.node.text
            matches = re.findall(regex, text, re.DOTALL)
            for match in matches:
                info["align"] += 1
                info["found_math"] = True
                match = replacement_manager.add_replacement(match, tag="math")
                text.replace(match, match)
            element.node.text = text

    def end_callback(element):
        pass

    body = tree.document.query_selector("body")
    traverse_dom(body, start_callback, end_callback)

    # Find any \equation environments
    def start_callback(element):
        regex = r"\\begin{equation}(.*?)\\end{equation}"
        if element.node.type == 3:
            text = element.node.text
            matches = re.findall(regex, text, re.DOTALL)
            for match in matches:
                info["equation"] += 1
                info["found_math"] = True
                match = match.replace("\\begin{equation}", "")
                match = match.replace("\\end{equation}", "")
                wrapped_text = wrap_math(match, display=True)
                wrapped_text = replacement_manager.add_replacement(
                    wrapped_text, tag="math"
                )
                text = text.replace(match, wrapped_text)
            # Remove the \begin{equation} and \end{equation} tags
            text = text.replace("\\begin{equation}", "")
            text = text.replace("\\end{equation}", "")
            element.node.text = text

    def end_callback(element):
        pass

    body = tree.document.query_selector("body")
    traverse_dom(body, start_callback, end_callback)

    # Find all .texerror
    texerrors = tree.document.query_selector_all(".texerror")
    for texerror in texerrors:
        # Find the text between {} (maximum length) and replace the texerror with that text
        match = re.search(r"\{(.{1,})\}", texerror.text)
        if match:
            info["found_math"] = True
            info["texerror"] += 1
            wrapped_match = wrap_math(match.group(1))
            texerror.html = replacement_manager.add_replacement(
                wrapped_match, tag="math"
            )

    # This has a ton of repeated code, but it's nice to have fine control over
    # how each source is handled.
    imgs = tree.document.query_selector_all("img")
    for img in imgs:
        class_attr = img.getattr("class")
        if class_attr is not None:
            class_list = class_attr.split(" ")
            if any([img_class in class_list for img_class in latex_image_class_names]):
                alt = img.getattr("alt")
                if alt is None:
                    continue
                update_math_parent_with_span(tree, img, alt, replacement_manager, info)
                info["img_math"] += 1

        src = img.getattr("src")
        if src is None:
            continue
        if "codecogs.com" in src:
            try:
                latex = src.split("?")[1:]
                latex = "?".join(latex)  # In case there are multiple ? in the latex
                latex = unquote(latex)
                update_math_parent_with_span(tree, img, latex, replacement_manager, info)
                info["codecogs_latex"] += 1
            except:
                pass
        if "latex.php" in src:
            try:
                # they usually have "alt='-i u_t + &#92;Delta u = |u|^2 u'"
                alt = img.getattr("alt")
                if alt is None:
                    continue
                # Unescape the latex
                alt = unquote(alt)
                # Get the latex
                update_math_parent_with_span(tree, img, alt, replacement_manager, info)
                info["wp_latex"] += 1
            except:
                pass
        if "/images/math/codecogs" in src:
            try:
                # they usually have "alt='-i u_t + &#92;Delta u = |u|^2 u'"
                alt = img.getattr("alt")
                if alt is None:
                    continue
                # Unescape the latex
                alt = unquote(alt)
                # Get the latex
                update_math_parent_with_span(tree, img, alt, replacement_manager, info)
                info["/images/math/codecogs"] += 1
            except:
                pass
        if "mimetex.cgi" in src:
            try:
                latex = src.split("?")[1:]
                latex = "?".join(latex)  # In case there are multiple ? in the latex
                latex = unquote(latex)
                update_math_parent_with_span(tree, img, latex, replacement_manager, info)
                info["mimetex.cgi"] += 1
            except:
                pass
        if "mathtex.cgi" in src:
            try:
                latex = src.split("?")[1:]
                latex = "?".join(latex)  # In case there are multiple ? in the latex
                latex = unquote(latex)
                update_math_parent_with_span(tree, img, latex, replacement_manager, info)
                info["mathtex.cgi"] += 1
            except:
                pass
        for latex_signal in ["tex?", "?latex=", "?tex="]:
            if latex_signal in src:
                try:
                    latex = src.split("latex_signal", 1)[1]
                    latex = unquote(latex)
                    update_math_parent_with_span(tree, img, latex, replacement_manager, info)
                    info["other_latex_img"] += 1
                except:
                    pass
        for latex_signal in ["tex", "latex"]:
            try:
                # they usually have "alt='-i u_t + &#92;Delta u = |u|^2 u'"
                alt = img.getattr(latex_signal)
                if alt is None:
                    continue
                # Unescape the latex
                alt = unquote(alt)
                # Get the latex
                update_math_parent_with_span(tree, img, alt, replacement_manager, info)
                info["other_latex_img"] += 1
            except:
                pass
        class_attr = img.getattr("class")
        if class_attr is not None:
            if "x-ck12" in class_attr:
                try:
                    latex = img.getattr("alt")
                    update_math_parent_with_span(tree, img, latex, replacement_manager, info)
                    info["x-ck12"] += 1
                except:
                    pass

    # Find any blocks with class math-container and replace them with spans
    math_containers = tree.document.query_selector_all(".math-container")
    for math_container in math_containers:
        text = math_container.text
        update_math_parent_with_span(tree, math_container, text, replacement_manager, info, display=True)
        info["math-container"] += 1

    katex_inline_wp = tree.document.query_selector_all(".wp-katex-eq")
    for katex in katex_inline_wp:
        text = katex.text
        display_attr = katex.getattr("data-display")
        if display_attr is not None:
            display = display_attr == "true"
        else:
            display = False
        update_math_parent_with_span(tree, katex, text, replacement_manager, info, display)
        info["wp-katex-eq"] += 1

    # Find all script[type="math/tex"] tags and replace them with spans
    latex_script_tags = tree.document.query_selector_all('script[type*="math/tex"]')
    for script_tag in latex_script_tags:
        text = script_tag.text
        mathjax_id = script_tag.getattr("id")
        display = "display" in script_tag.getattr("type")
        math_parent = update_math_parent_with_span(tree, script_tag, text, replacement_manager, info, display)
        if mathjax_id:
            mathjax_id = "-".join(mathjax_id.split("-")[:3])
            for unused_tag in math_parent.query_selector_all(f'[id*="{mathjax_id}"]'):
                unused_tag_parent = unused_tag.parent
                unused_tag_parent.remove_child(unused_tag)
        info["script_math_tex"] += 1

    asciimath_script_tags = tree.document.query_selector_all(
        'script[type*="math/asciimath"]'
    )
    for script_tag in asciimath_script_tags:
        try:
            text = script_tag.text
            text = extract_asciimath(text)
            update_math_parent_with_span(tree, script_tag, text, replacement_manager, info)
            info["script_math_asciimath"] += 1
        except:
            # Delete this script tag
            parent = script_tag.parent
            parent.remove_child(script_tag)

    mathml_script_tags = tree.document.query_selector_all('script[type*="math/mml"]')
    for script_tag in mathml_script_tags:
        try:
            # Try translating to LaTeX
            mathml = script_tag.text
            mathml = html_unescape(mathml)
            # If this includes xmlns:mml, then we need to replace all
            # instances of mml: with nothing
            if "xmlns:mml" in mathml:
                mathml = mathml.replace("mml:", "")
                # replace xmlns:mml="..." with nothing
                mathml = re.sub(r'xmlns:mml=".*?"', "", mathml)
            latex = mml_to_latex(mathml)
            update_math_parent_with_span(tree, script_tag, latex, replacement_manager, info)
            info["mathml"] += 1
        except Exception as e:
            # Delete this script tag
            parent = script_tag.parent
            parent.remove_child(script_tag)

    for tex_attr in ["tex", "data-tex", "data-formula"]:
        for tex_attr_tag in tree.document.query_selector_all(f"[{tex_attr}]"):
            try:
                text = tex_attr_tag.getattr(tex_attr)
                if text is None:
                    continue
                text = html_unescape(unquote(text))
                update_math_parent_with_span(tree, tex_attr_tag, text, replacement_manager, info)
            except:
                pass

    for tex_attr in ["mathml", "data-mathml"]:
        for tex_attr_tag in tree.document.query_selector_all(f"[{tex_attr}]"):
            try:
                mathml = tex_attr_tag.getattr(tex_attr)
                if mathml is None:
                    continue
                mathml = html_unescape(mathml)
                # If this includes xmlns:mml, then we need to replace all
                # instances of mml: with nothing
                if "xmlns:mml" in mathml:
                    mathml = mathml.replace("mml:", "")
                    # replace xmlns:mml="..." with nothing
                    mathml = re.sub(r'xmlns:mml=".*?"', "", mathml)
                latex = mml_to_latex(mathml)
                update_math_parent_with_span(tree, tex_attr_tag, latex, replacement_manager, info)
                info["mathml"] += 1
            except Exception as e:
                # Delete this script tag
                parent = tex_attr_tag.parent
                parent.remove_child(tex_attr_tag)

    # For katex, find all elements with class = tex
    katex_spans = tree.document.query_selector_all(".tex")
    for katex_span in katex_spans:
        try:
            # Check if they have data-expr attr
            expr = katex_span.getattr("data-expr")
            if expr is None:
                continue
            # Replace with a span
            update_math_parent_with_span(tree, katex_span, expr, replacement_manager, info)
            info["katex"] += 1
        except:
            pass

    # For mathhelpforum.com
    formula_spans = tree.document.query_selector_all('formula[class*="math"]')
    for formula_span in formula_spans:
        try:
            expr = formula_span.text
            if expr.strip():
                update_math_parent_with_span(
                    tree, formula_span, expr, replacement_manager, info
                )
                info["katex"] += 1
        except:
            pass

    # Find any spans with class "katex"
    katex_spans = tree.document.query_selector_all("span.katex")
    for katex_span in katex_spans:
        # Find any spans with class "katex-html" and remove them
        katex_html_spans = katex_span.query_selector_all("span.katex-html")
        for katex_html_span in katex_html_spans:
            parent = katex_html_span.parent
            parent.remove_child(katex_html_span)

    # Remove any .MathJax_Preview spans
    mathjax_preview_spans = tree.document.query_selector_all("span.MathJax_Preview")
    for mathjax_preview_span in mathjax_preview_spans:
        parent = mathjax_preview_span.parent
        parent.remove_child(mathjax_preview_span)

    # Find any math tags
    math_tags = tree.document.query_selector_all("math")
    # For each math tag, see if there is an annotation tag with
    # encoding="application/x-tex" inside it
    for math_tag in math_tags:
        annotation_tag = math_tag.query_selector(
            'annotation[encoding="application/x-tex"]'
        )
        if annotation_tag is not None:
            # Get the text content of the annotation tag
            text = annotation_tag.text
            math_parent = update_math_parent_with_span(tree, math_tag, text, replacement_manager, info)
            # If the parent has style="display:none", then we need to
            # remove the style attribute
            style_value = math_parent.getattr("style")
            if style_value is not None:
                normalized_style_value = (
                    style_value.lower().strip().replace(" ", "").replace(";", "")
                )
                if "display:none" in normalized_style_value:
                    math_parent.delattr("style")
            info["math_annotations"] += 1
        # Check if the math tag has an alttext attribute
        elif (
            math_tag.getattr("alttext") is not None
            or math_tag.getattr("data-code") is not None
        ):
            # Get the alttext attribute
            if math_tag.getattr("alttext") is not None:
                alttext = math_tag.getattr("alttext")
            else:
                alttext = math_tag.getattr("data-code")
            alttext = html_unescape(unquote(alttext))
            # Set the html of the new span tag to the text
            if math_tag.getattr("data-math-language") == "asciimath":
                alttext = extract_asciimath(alttext)
            elif math_tag.getattr("data-math-language") == "mathml":
                try:
                    alttext = mml_to_latex(alttext)
                except:
                    pass
            else:
                pass

            update_math_parent_with_span(tree, math_tag, alttext, replacement_manager, info)
            info["math_alttext"] += 1
        # Otherwise, translate the math tag to LaTeX
        else:
            try:
                # Try translating to LaTeX
                mathml = math_tag.html
                mathml = html_unescape(mathml)
                # If this includes xmlns:mml, then we need to replace all
                # instances of mml: with nothing
                try:
                    if "xmlns:mml" in mathml:
                        mathml = mathml.replace("mml:", "")
                        # replace xmlns:mml="..." with nothing
                        mathml = re.sub(r'xmlns:mml=".*?"', "", mathml)
                    latex = mml_to_latex(mathml)
                except:
                    latex = fast_html2text(mathml)

                update_math_parent_with_span(
                    tree, math_tag, latex, replacement_manager, info
                )
                info["mathml"] += 1

            except Exception as e:
                parent = math_tag.parent
                parent.remove_child(math_tag)

    mathjax_tags = tree.document.query_selector_all("mathjax")
    for mathjax_tag in mathjax_tags:
        # Get the inner text of the mathjax tag
        text = mathjax_tag.text
        text = html_unescape(text)
        # Use regex to find text wrapped in hashes
        matches = re.findall(r"#(.+?)#", text)
        # For each match, replace the match with the LaTeX
        for match in matches:
            try:
                latex = extract_asciimath(match)
                # Replace the match with the LaTeX
                text = text.replace(f"#{match}#", latex)
            except Exception as e:
                pass

        update_math_parent_with_span(tree, mathjax_tag, text, replacement_manager, info)


def remove_color(text):
    return re.sub(color_regex, "", text)



"""A suite of tools for working with highlight.js."""

import contextlib
import json
import os
import pathlib
import shutil
import subprocess
import tempfile
import textwrap
import typing


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


HLJS_VERSION = "11.9.0"
HLJS_REPO = "https://githubfast.com/highlightjs/highlight.js"


def delete_old_hljs_scripts() -> None:
    """Deletes all JS files associated with HLJS in assets/."""
    for f in os.listdir("assets"):
        if f.startswith("_ch-hljs-lang-"):
            os.remove(f"assets/{f}")
    os.remove("assets/_ch-highlight.js")


class HLJSBuild(typing.NamedTuple):
    hljs_file: pathlib.Path
    hljs_min_file: pathlib.Path


def build_hljs() -> HLJSBuild:
    """Builds highlight.js in the current directory."""
    hljs_dir = "highlight.js"
    subprocess.run(["git", "clone", HLJS_REPO])
    with pushd(hljs_dir):
        subprocess.run(["git", "checkout", HLJS_VERSION])
        subprocess.run(["npm", "install"])
        subprocess.run(["node", "tools/build.js", "-t", "browser"])

        hljs_file = "build/highlight.js"
    return HLJSBuild(
        hljs_file=pathlib.Path(f"{hljs_dir}/build/highlight.js"),
        hljs_min_file=pathlib.Path(f"{hljs_dir}/build/highlight.min.js"),
    )


class Language(typing.NamedTuple):
    name: str
    alias: str


def extract_languages_from_highlight_js(
    highlight_js_path: pathlib.Path,
) -> typing.Iterator[Language]:
    """Extracts languages supported by highlight.js."""
    extraction = subprocess.run(
        ["node", "tools/extract-hljs-languages.js", highlight_js_path.resolve()],
        capture_output=True,
        check=True,
    )
    for line in extraction.stdout.decode("utf-8").splitlines():
        language_json = json.loads(line)
        yield Language(name=language_json["name"], alias=language_json["alias"])


def generate_hljs_languages_python_list(
    languages: typing.Iterable[Language], out: typing.TextIO
):
    """Generates a Python module containing a list of all supported languages."""
    out.write(
        textwrap.dedent(
            f"""\
      \"""A list of all supported languages in highlight.js.

      This file is automatically generated by tools.
      \"""
      import typing


      class Language(typing.NamedTuple):
          \"""A language supported by highlight.js.\"""
          name: str
          alias: str


      languages: typing.List[Language] = [
    """
        )
    )

    for language in languages:
        out.write(f"    {repr(language)},\n")
    out.write("]")


def update_hljs_main():
    """Installs HLJS_VERSION of highlight.js."""
    with tempfile.TemporaryDirectory() as d:
        with pushd(d):
            hljs_build = build_hljs()
        hljs_build = HLJSBuild(
            hljs_file=pathlib.Path(d) / hljs_build.hljs_file,
            hljs_min_file=pathlib.Path(d) / hljs_build.hljs_min_file,
        )

        with open("dictionary/highlightjs.py", "w") as hljslangs_py:
            generate_hljs_languages_python_list(
                extract_languages_from_highlight_js(hljs_build.hljs_file), hljslangs_py
            )


if __name__ == "__main__":
    update_hljs_main()



# pylint:disable-msg=E0611,I1101
"""Minimalistic fork of readability-lxml code

This is a python port of a ruby port of arc90's readability project

http://lab.arc90.com/experiments/readability/

Given a html document, it pulls out the main body text and cleans it up.

Ruby port by starrhorne and iterationlabs
Python port by gfxmonk

For list of contributors see
https://github.com/timbertson/python-readability
https://github.com/buriy/python-readability

License of forked code: Apache-2.0.
"""

import logging
import re

from math import sqrt
from operator import attrgetter

from lxml.etree import tostring
from lxml.html import fragment_fromstring

from .utils import load_html, trim

LOGGER = logging.getLogger(__name__)


DOT_SPACE = re.compile(r"\.( |$)")


def _tostring(string):
    return tostring(string, encoding=str, method="xml")


DIV_TO_P_ELEMS = {
    "a",
    "blockquote",
    "dl",
    "div",
    "img",
    "ol",
    "p",
    "pre",
    "table",
    "ul",
}

DIV_SCORES = {"div", "article"}
BLOCK_SCORES = {"pre", "td", "blockquote"}
BAD_ELEM_SCORES = {"address", "ol", "ul", "dl", "dd", "dt", "li", "form", "aside"}
STRUCTURE_SCORES = {"h1", "h2", "h3", "h4", "h5", "h6", "th", "header", "footer", "nav"}

TEXT_CLEAN_ELEMS = {"p", "img", "li", "a", "embed", "input"}

REGEXES = {
    "unlikelyCandidatesRe": re.compile(
        r"combx|comment|community|disqus|extra|foot|header|menu|remark|rss|shoutbox|sidebar|sponsor|ad-break|agegate|pagination|pager|popup|tweet|twitter",
        re.I,
    ),
    "okMaybeItsACandidateRe": re.compile(r"and|article|body|column|main|shadow", re.I),
    "positiveRe": re.compile(
        r"article|body|content|entry|hentry|main|page|pagination|post|text|blog|story",
        re.I,
    ),
    "negativeRe": re.compile(
        r"button|combx|comment|com-|contact|figure|foot|footer|footnote|form|input|masthead|media|meta|outbrain|promo|related|scroll|shoutbox|sidebar|sponsor|shopping|tags|tool|widget",
        re.I,
    ),
    "divToPElementsRe": re.compile(
        r"<(?:a|blockquote|dl|div|img|ol|p|pre|table|ul)", re.I
    ),
    "videoRe": re.compile(r"https?:\/\/(?:www\.)?(?:youtube|vimeo)\.com", re.I),
}

FRAME_TAGS = {"body", "html"}
LIST_TAGS = {"ol", "ul"}
# DIV_TO_P_ELEMS = {'a', 'blockquote', 'dl', 'div', 'img', 'ol', 'p', 'pre', 'table', 'ul'}


def text_length(elem):
    "Return the length of the element with all its contents."
    return len(trim(elem.text_content()))


class Candidate:
    "Defines a class to score candidate elements."

    __slots__ = ["score", "elem"]

    def __init__(self, score, elem):
        self.score = score
        self.elem = elem


class Document:
    """Class to build a etree document out of html."""

    __slots__ = ["doc", "min_text_length", "retry_length"]

    def __init__(self, doc, min_text_length=25, retry_length=250):
        """Generate the document

        :param doc: string of the html content.
        :param min_text_length: Set to a higher value for more precise detection of longer texts.
        :param retry_length: Set to a lower value for better detection of very small texts.

        The Document class is not re-enterable.
        It is designed to create a new Document() for each HTML file to process it.

        API method:
        .summary() -- cleaned up content
        """
        self.doc = doc
        self.min_text_length = min_text_length
        self.retry_length = retry_length

    def summary(self):
        """
        Given a HTML file, extracts the text of the article.

        Warning: It mutates internal DOM representation of the HTML document,
        so it is better to call other API methods before this one.
        """
        for elem in self.doc.iter("script", "style"):
            elem.drop_tree()

        ruthless = True
        while True:
            if ruthless:
                self.remove_unlikely_candidates()
            self.transform_misused_divs_into_paragraphs()
            candidates = self.score_paragraphs()

            best_candidate = self.select_best_candidate(candidates)

            if best_candidate:
                article = self.get_article(candidates, best_candidate)
            else:
                if ruthless is True:
                    ruthless = False
                    LOGGER.debug(
                        "Ended up stripping too much - going for a safer parse"
                    )
                    # try again
                    continue
                # go ahead
                LOGGER.debug(
                    "Ruthless and lenient parsing did not work. Returning raw html"
                )
                article = self.doc.find("body")
                if article is None:
                    article = self.doc

            cleaned_article = self.sanitize(article, candidates)
            article_length = len(cleaned_article or "")
            if ruthless and article_length < self.retry_length:
                ruthless = False
                # Loop through and try again.
                continue
            return cleaned_article

    def get_article(self, candidates, best_candidate):
        # Now that we have the top candidate, look through its siblings for
        # content that might also be related.
        # Things like preambles, content split by ads that we removed, etc.
        sibling_score_threshold = max(10, best_candidate.score * 0.2)
        # create a new html document with a div
        output = fragment_fromstring("<div/>")
        parent = best_candidate.elem.getparent()
        siblings = parent.getchildren() if parent is not None else [best_candidate.elem]
        for sibling in siblings:
            # in lxml there no concept of simple text
            # if isinstance(sibling, NavigableString): continue
            append = False
            # conditions
            if sibling == best_candidate.elem or (
                sibling in candidates
                and candidates[sibling].score >= sibling_score_threshold
            ):
                append = True
            elif sibling.tag == "p":
                link_density = self.get_link_density(sibling)
                node_content = sibling.text or ""
                node_length = len(node_content)

                if (
                    node_length > 80
                    and link_density < 0.25
                    or (
                        node_length <= 80
                        and link_density == 0
                        and DOT_SPACE.search(node_content)
                    )
                ):
                    append = True
            # append to the output div
            if append:
                output.append(sibling)
        # if output is not None:
        #    output.append(best_candidate.elem)
        return output

    def select_best_candidate(self, candidates):
        if not candidates:
            return None
        sorted_candidates = sorted(
            candidates.values(), key=attrgetter("score"), reverse=True
        )
        if LOGGER.isEnabledFor(logging.DEBUG):
            for candidate in sorted_candidates[:5]:
                LOGGER.debug("Top 5: %s %s", candidate.elem.tag, candidate.score)
        return next(iter(sorted_candidates))

    def get_link_density(self, elem):
        total_length = text_length(elem) or 1
        link_length = sum(text_length(link) for link in elem.findall(".//a"))
        return link_length / total_length

    def score_paragraphs(self):
        candidates = {}

        for elem in self.doc.iter("p", "pre", "td"):
            parent_node = elem.getparent()
            if parent_node is None:
                continue
            grand_parent_node = parent_node.getparent()

            elem_text = trim(elem.text_content())
            elem_text_len = len(elem_text)

            # discard too short paragraphs
            if elem_text_len < self.min_text_length:
                continue

            for node in (parent_node, grand_parent_node):
                if node is not None and node not in candidates:
                    candidates[node] = self.score_node(node)

            score = 1 + len(elem_text.split(",")) + min((elem_text_len / 100), 3)
            # if elem not in candidates:
            #    candidates[elem] = self.score_node(elem)

            candidates[parent_node].score += score
            if grand_parent_node is not None:
                candidates[grand_parent_node].score += score / 2

        # Scale the final candidates score based on link density. Good content
        # should have a relatively small link density (5% or less) and be
        # mostly unaffected by this operation.
        for elem, candidate in candidates.items():
            candidate.score *= 1 - self.get_link_density(elem)

        return candidates

    def class_weight(self, elem):
        weight = 0
        for attribute in filter(None, (elem.get("class"), elem.get("id"))):
            if REGEXES["negativeRe"].search(attribute):
                weight -= 25
            if REGEXES["positiveRe"].search(attribute):
                weight += 25
        return weight

    def score_node(self, elem):
        score = self.class_weight(elem)
        name = elem.tag.lower()
        if name in DIV_SCORES:
            score += 5
        elif name in BLOCK_SCORES:
            score += 3
        elif name in BAD_ELEM_SCORES:
            score -= 3
        elif name in STRUCTURE_SCORES:
            score -= 5
        return Candidate(score, elem)

    def remove_unlikely_candidates(self):
        for elem in self.doc.findall(".//*"):
            attrs = " ".join(filter(None, (elem.get("class"), elem.get("id"))))
            if len(attrs) < 2:
                continue
            if (
                elem.tag not in FRAME_TAGS
                and REGEXES["unlikelyCandidatesRe"].search(attrs)
                and (not REGEXES["okMaybeItsACandidateRe"].search(attrs))
            ):
                # LOGGER.debug("Removing unlikely candidate: %s", elem.tag)
                elem.drop_tree()

    def transform_misused_divs_into_paragraphs(self):
        for elem in self.doc.findall(".//div"):
            # transform <div>s that do not contain other block elements into
            # <p>s
            # FIXME: The current implementation ignores all descendants that
            # are not direct children of elem
            # This results in incorrect results in case there is an <img>
            # buried within an <a> for example
            # hurts precision:
            # if not any(e.tag in DIV_TO_P_ELEMS for e in list(elem)):
            if not REGEXES["divToPElementsRe"].search(
                "".join(map(_tostring, list(elem)))
            ):
                elem.tag = "p"

        for elem in self.doc.findall(".//div"):
            if elem.text and elem.text.strip():
                p_elem = fragment_fromstring("<p/>")
                p_elem.text, elem.text = elem.text, None
                elem.insert(0, p_elem)

            for pos, child in sorted(enumerate(elem), reverse=True):
                if child.tail and child.tail.strip():
                    p_elem = fragment_fromstring("<p/>")
                    p_elem.text, child.tail = child.tail, None
                    elem.insert(pos + 1, p_elem)
                if child.tag == "br":
                    child.drop_tree()

    def sanitize(self, node, candidates):
        for header in node.iter("h1", "h2", "h3", "h4", "h5", "h6"):
            if self.class_weight(header) < 0 or self.get_link_density(header) > 0.33:
                header.drop_tree()

        for elem in node.iter("form", "textarea"):
            elem.drop_tree()

        for elem in node.iter("iframe"):
            if "src" in elem.attrib and REGEXES["videoRe"].search(elem.attrib["src"]):
                elem.text = "VIDEO"  # ADD content to iframe text node to force <iframe></iframe> proper output
            else:
                elem.drop_tree()

        allowed = set()
        # Conditionally clean <table>s, <ul>s, and <div>s
        for elem in reversed(
            node.xpath("//table|//ul|//div|//aside|//header|//footer|//section")
        ):
            if elem in allowed:
                continue
            weight = self.class_weight(elem)
            score = candidates[elem].score if elem in candidates else 0
            if weight + score < 0:
                LOGGER.debug(
                    "Removed %s with score %6.3f and weight %-3s",
                    elem.tag,
                    score,
                    weight,
                )
                elem.drop_tree()
            elif elem.text_content().count(",") < 10:
                to_remove = False
                counts = {
                    kind: len(elem.findall(f".//{kind}")) for kind in TEXT_CLEAN_ELEMS
                }
                counts["li"] -= 100
                counts["input"] -= len(elem.findall('.//input[@type="hidden"]'))

                # Count the text length excluding any surrounding whitespace
                content_length = text_length(elem)
                link_density = self.get_link_density(elem)
                parent_node = elem.getparent()
                if parent_node is not None:
                    score = (
                        candidates[parent_node].score
                        if parent_node in candidates
                        else 0
                    )
                # if elem.tag == 'div' and counts["img"] >= 1:
                #    continue
                if counts["p"] and counts["img"] > 1 + counts["p"] * 1.3:
                    reason = f'too many images ({counts["img"]})'
                    to_remove = True
                elif counts["li"] > counts["p"] and elem.tag not in LIST_TAGS:
                    reason = "more <li>s than <p>s"
                    to_remove = True
                elif counts["input"] > (counts["p"] / 3):
                    reason = "less than 3x <p>s than <input>s"
                    to_remove = True
                elif content_length < self.min_text_length and counts["img"] == 0:
                    reason = f"too short content length {content_length} without a single image"
                    to_remove = True
                elif content_length < self.min_text_length and counts["img"] > 2:
                    reason = (
                        f"too short content length {content_length} and too many images"
                    )
                    to_remove = True
                elif weight < 25 and link_density > 0.2:
                    reason = (
                        f"too many links {link_density:.3f} for its weight {weight}"
                    )
                    to_remove = True
                elif weight >= 25 and link_density > 0.5:
                    reason = (
                        f"too many links {link_density:.3f} for its weight {weight}"
                    )
                    to_remove = True
                elif (counts["embed"] == 1 and content_length < 75) or counts[
                    "embed"
                ] > 1:
                    reason = (
                        "<embed>s with too short content length, or too many <embed>s"
                    )
                    to_remove = True
                elif not content_length:
                    reason = "no content"
                    to_remove = True

                    # find x non empty preceding and succeeding siblings
                    siblings = []
                    for sib in elem.itersiblings():
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            siblings.append(sib_content_length)
                            # if len(siblings) >= 1:
                            break
                    limit = len(siblings) + 1
                    for sib in elem.itersiblings(preceding=True):
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            siblings.append(sib_content_length)
                            if len(siblings) >= limit:
                                break
                    if siblings and sum(siblings) > 1000:
                        to_remove = False
                        allowed.update(elem.iter("table", "ul", "div", "section"))

                if to_remove:
                    elem.drop_tree()
                    if LOGGER.isEnabledFor(logging.DEBUG):
                        LOGGER.debug(
                            "Removed %6.3f %s with weight %s cause it has %s.",
                            score,
                            elem.tag,
                            weight,
                            reason or "",
                        )

        self.doc = node
        return _tostring(self.doc)


"""
Port of isProbablyReaderable from mozilla/readability.js to Python.

https://github.com/mozilla/readability

License of forked code: Apache-2.0.
"""

REGEXPS = {
    "unlikelyCandidates": re.compile(
        r"-ad-|ai2html|banner|breadcrumbs|combx|comment|community|cover-wrap|disqus|extra|footer|gdpr|header|legends|menu|related|remark|replies|rss|shoutbox|sidebar|skyscraper|social|sponsor|supplemental|ad-break|agegate|pagination|pager|popup|yom-remote",
        re.I,
    ),
    "okMaybeItsACandidate": re.compile(
        r"and|article|body|column|content|main|shadow", re.I
    ),
}

DISPLAY_NONE = re.compile(r"display:\s*none", re.I)


def is_node_visible(node):
    """
    Checks if the node is visible by considering style, attributes, and class.
    """

    if "style" in node.attrib and DISPLAY_NONE.search(node.get("style")):
        return False
    if "hidden" in node.attrib:
        return False
    if node.get("aria-hidden") == "true" and "fallback-image" not in node.get(
        "class", ""
    ):
        return False
    return True


def is_probably_readerable(html, options={}):
    """
    Decides whether or not the document is reader-able without parsing the whole thing.
    """
    doc = load_html(html)

    min_content_length = options.get("min_content_length", 140)
    min_score = options.get("min_score", 20)
    visibility_checker = options.get("visibility_checker", is_node_visible)

    nodes = set(doc.xpath(".//p | .//pre | .//article"))
    nodes.update(node.getparent() for node in doc.xpath(".//div/br"))

    score = 0
    for node in nodes:
        if not visibility_checker(node):
            continue

        class_and_id = f"{node.get('class', '')} {node.get('id', '')}"
        if REGEXPS["unlikelyCandidates"].search(class_and_id) and not REGEXPS[
            "okMaybeItsACandidate"
        ].search(class_and_id):
            continue

        if node.xpath("./parent::li/p"):
            continue

        text_content_length = len(node.text_content().strip())
        if text_content_length < min_content_length:
            continue

        score += sqrt(text_content_length - min_content_length)
        if score > min_score:
            return True

    return False



"""Handle the <img> tag."""
from typing import Dict

from haruka_parser.table_processing.html_state import HtmlDocumentState


def img_start_handler(state: HtmlDocumentState, attrs: Dict) -> None:
    """Handle the <img> tag."""
    image_text = attrs.get("alt", "") or attrs.get("title", "")
    if image_text and not (
        state.config.deduplicate_captions and image_text == state.last_caption
    ):
        state.tags[-1].write(f"[{image_text}]")
        state.last_caption = image_text



"""
This script implements a basic guided user interface (GUI).
"""

import sys

from gooey import Gooey, GooeyParser

from . import __version__
from .cli import add_args, process_args


DESCRIPTION = 'Web scraping tool for text discovery and extraction'


@Gooey(
    program_name='Trafilatura Graphical User Interface',
    default_size=(840, 720),
    tabbed_groups=True,
    required_cols=1,
    optional_cols=1,
    run_validators=True,
    clear_before_run=True,
    menu=[{
        'name': 'File',
        'items': [{
                'type': 'AboutDialog',
                'menuTitle': 'About',
                'name': 'Trafilatura-GUI',
                'description': DESCRIPTION,
                'version': __version__,
                'copyright': '2021',
                'website': 'https://trafilatura.readthedocs.io',
                'developer': 'Adrien Barbaresi',
                'license': "Apache 2.0"
            }, {
                'type': 'Link',
                'menuTitle': 'Visit the website',
                'url': 'https://trafilatura.readthedocs.io'
            }]
        },{
        'name': 'Help',
        'items': [{
            'type': 'Link',
            'menuTitle': 'Documentation',
            'url': 'https://trafilatura.readthedocs.io/'
        }]
    }]
)


def main():
    "Run as a Gooey/GUI utility."
    parser = GooeyParser(description=DESCRIPTION)
    parser = add_args(parser)

    # https://github.com/chriskiehl/Gooey/blob/master/docs/Gooey-Options.md
    args = parser.parse_args(sys.argv[1:])
    # safety: no input
    if not args.URL and not args.inputfile and not args.inputdir:
        sys.exit("Missing input, exiting.")
    # process
    process_args(args)


if __name__ == '__main__':
    main()



"""
Scrapes the main text of web pages while preserving some structure
https://github.com/adbar/trafilatura
"""

import re
from pathlib import Path

from setuptools import setup


def get_version(package):
    "Return package version as listed in `__version__` in `init.py`"
    initfile = Path(package, "__init__.py").read_text()  # Python >= 3.5
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", initfile)[1]


def get_long_description():
    "Return the README"
    with open("README.md", "r", encoding="utf-8") as filehandle:
        long_description = filehandle.read()
    return long_description


# some problems with installation solved this way
extras = {
    "all": [
        "brotli",
        "cchardet >= 2.1.7; python_version < '3.11'",  # build issue
        "faust-cchardet >= 2.1.19; python_version >= '3.11'",
        "htmldate[speed] >= 1.8.1",
        "py3langid >= 0.2.2",
        "pycurl >= 7.45.3",
        "zstandard >= 0.20.0",
    ],
    "gui": [
        "Gooey >= 1.0.1",
    ],
}

setup(
    name="trafilatura",
    version=get_version("trafilatura"),
    description="Python package and command-line tool designed to gather text on the Web, includes all necessary discovery and text processing components to perform web crawling, downloads, scraping, and extraction of main texts, metadata and comments.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        # As from https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        #'Development Status :: 6 - Mature',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Security",
        "Topic :: Text Editors :: Text Processing",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: Markdown",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Utilities",
    ],
    keywords=[
        "corpus",
        "html2text",
        "news-crawler",
        "natural-language-processing",
        "scraper",
        "tei-xml",
        "text-extraction",
        "webscraping",
        "web-scraping",
    ],
    url="https://trafilatura.readthedocs.io",
    project_urls={
        "Documentation": "https://trafilatura.readthedocs.io",
        "Source": "https://github.com/adbar/trafilatura",
        "Blog": "https://adrien.barbaresi.eu/blog/tag/trafilatura.html",
    },
    author="Adrien Barbaresi",
    author_email="barbaresi@bbaw.de",
    license="Apache-2.0",
    packages=["trafilatura"],
    package_data={
        "trafilatura": [
            "data/tei-schema-pickle.lzma",
            "data/jt-stopwords-pickle.lzma",
            "settings.cfg",
        ]
    },
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "certifi",
        "charset_normalizer >= 3.0.1; python_version < '3.7'",
        "charset_normalizer >= 3.2.0; python_version >= '3.7'",
        "courlan >= 1.1.0",
        "htmldate >= 1.8.1",
        "importlib_metadata; python_version < '3.8'",
        "justext >= 3.0.1",
        # see tests on Github Actions
        "lxml == 4.9.2 ; platform_system == 'Darwin' and python_version <= '3.8'",
        "lxml >= 5.2.2 ; platform_system != 'Darwin' or python_version > '3.8'",
        "urllib3 >= 1.26, < 2; python_version < '3.7'",
        "urllib3 >= 1.26, < 3; python_version >= '3.7'",
    ],
    extras_require=extras,
    entry_points={
        "console_scripts": [
            "trafilatura=trafilatura.cli:main",
            "trafilatura_gui=trafilatura.gui:main",
        ],
    },
    # platforms='any',
    tests_require=["pytest"],
    zip_safe=False,
)



# pylint:disable-msg=E0611,I1101
"""
All functions needed to steer and execute downloads of web documents.
"""

import logging
import random

from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from io import BytesIO
from time import sleep

import certifi
import urllib3

try:
    import pycurl
    CURL_SHARE = pycurl.CurlShare()
    # available options:
    # https://curl.se/libcurl/c/curl_share_setopt.html
    CURL_SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_DNS)
    CURL_SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_SSL_SESSION)
    # not thread-safe
    # CURL_SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_CONNECT)
except ImportError:
    pycurl = None

from courlan import UrlStore
from courlan.network import redirection_test

try:  # Python 3.8+
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

from .settings import DEFAULT_CONFIG, Extractor
from .utils import URL_BLACKLIST_REGEX, decode_file, is_acceptable_length, make_chunks


LOGGER = logging.getLogger(__name__)

NUM_CONNECTIONS = 50

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
HTTP_POOL = None
NO_CERT_POOL = None
RETRY_STRATEGY = None

DEFAULT_HEADERS = urllib3.util.make_headers(accept_encoding=True)
USER_AGENT = 'trafilatura/' + version("trafilatura") + ' (+https://github.com/adbar/trafilatura)'
DEFAULT_HEADERS['User-Agent'] = USER_AGENT


class Response:
    "Store information gathered in a HTTP response object."
    __slots__ = ["data", "headers", "html", "status", "url"]

    def __init__(self, data, status, url):
        self.data = data
        self.headers = None
        self.html = None
        self.status = status
        self.url = url

    def __bool__(self):
        return self.data is not None

    def __repr__(self):
        return self.html if self.html else decode_file(self.data)

    def __str__(self):
        return self.__repr__()

    def store_headers(self, headerdict):
        "Store response headers if required."
        # further control steps here
        self.headers = {k.lower(): v for k, v in headerdict.items()}

    def decode_data(self, decode):
        "Decode the bytestring in data and store a string in html."
        if decode and self.data:
            self.html = decode_file(self.data)

    def as_dict(self):
        "Convert the response object to a dictionary."
        return {
            attr: getattr(self, attr)
            for attr in self.__slots__
            if hasattr(self, attr)
        }


# caching throws an error
# @lru_cache(maxsize=2)
def _parse_config(config):
    'Read and extract HTTP header strings from the configuration file.'
    # load a series of user-agents
    myagents = config.get('DEFAULT', 'USER_AGENTS').strip() or None
    if myagents is not None and myagents != '':
        myagents = myagents.split("\n")
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Cookies
    # todo: support for several cookies?
    mycookie = config.get('DEFAULT', 'COOKIE') or None
    return myagents, mycookie


def _determine_headers(config, headers=None):
    'Internal function to decide on user-agent string.'
    if config != DEFAULT_CONFIG:
        myagents, mycookie = _parse_config(config)
        headers = {}
        if myagents is not None:
            rnumber = random.randint(0, len(myagents) - 1)
            headers['User-Agent'] = myagents[rnumber]
        if mycookie is not None:
            headers['Cookie'] = mycookie
    return headers or DEFAULT_HEADERS


def _send_urllib_request(url, no_ssl, with_headers, config):
    "Internal function to robustly send a request (SSL or not) and return its result."
    # customize headers
    global HTTP_POOL, NO_CERT_POOL, RETRY_STRATEGY
    if not RETRY_STRATEGY:
        RETRY_STRATEGY = urllib3.util.Retry(
            total=config.getint("DEFAULT", "MAX_REDIRECTS"),
            redirect=config.getint("DEFAULT", "MAX_REDIRECTS"), # raise_on_redirect=False,
            connect=0,
            backoff_factor=config.getint('DEFAULT', 'DOWNLOAD_TIMEOUT')/2,
            status_forcelist=[
                429, 499, 500, 502, 503, 504, 509, 520, 521, 522, 523, 524, 525, 526, 527, 530, 598
            ],
            # unofficial: https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#Unofficial_codes
        )
    try:
        # TODO: read by streaming chunks (stream=True, iter_content=xx)
        # so we can stop downloading as soon as MAX_FILE_SIZE is reached
        if no_ssl is False:
            # define pool
            if not HTTP_POOL:
                HTTP_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY, timeout=config.getint('DEFAULT', 'DOWNLOAD_TIMEOUT'), ca_certs=certifi.where(), num_pools=NUM_CONNECTIONS)  # cert_reqs='CERT_REQUIRED'
            # execute request
            response = HTTP_POOL.request('GET', url, headers=_determine_headers(config), retries=RETRY_STRATEGY)
        else:
            # define pool
            if not NO_CERT_POOL:
                NO_CERT_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY, timeout=config.getint('DEFAULT', 'DOWNLOAD_TIMEOUT'), cert_reqs='CERT_NONE', num_pools=NUM_CONNECTIONS)
            # execute request
            response = NO_CERT_POOL.request('GET', url, headers=_determine_headers(config), retries=RETRY_STRATEGY)
    except urllib3.exceptions.SSLError:
        LOGGER.warning('retrying after SSLError: %s', url)
        return _send_urllib_request(url, True, with_headers, config)
    except Exception as err:
        LOGGER.error('download error: %s %s', url, err)  # sys.exc_info()[0]
    else:
        # necessary for standardization
        resp = Response(response.data, response.status, response.geturl())
        if with_headers:
            resp.store_headers(response.headers)
        return resp
    # catchall
    return None


def _handle_response(url, response, decode, options):
    'Internal function to run safety checks on response result.'
    lentest = len(response.html or response.data or "")
    if response.status != 200:
        LOGGER.error('not a 200 response: %s for URL %s', response.status, url)
    # raise error instead?
    elif is_acceptable_length(lentest, options):
        return response.html if decode else response
    # catchall
    return None


def fetch_url(url, decode=True, no_ssl=False, config=DEFAULT_CONFIG, options=None):
    """Downloads a web page and seamlessly decodes the response.

    Args:
        url: URL of the page to fetch.
        no_ssl: Don't try to establish a secure connection (to prevent SSLError).
        config: Pass configuration values for output control.
        options: Extraction options (supersedes config).

    Returns:
        Unicode string or None in case of failed downloads and invalid results.

    """
    if not decode:
        raise ValueError(
              "Raw response objects are deprecated for fetch_url(), use fetch_response() instead."
              )
    response = fetch_response(url, decode=decode, no_ssl=no_ssl, config=config)
    if response is not None and response != '':
        if not options:
            options = Extractor(config=config)
        return _handle_response(url, response, decode, options)
        # return '' (useful do discard further processing?)
        # return response
    return None


def fetch_response(url, *, decode=False, no_ssl=False, with_headers=False, config=DEFAULT_CONFIG):
    """Downloads a web page and returns a full response object.

    Args:
        url: URL of the page to fetch.
        decode: Use html attribute to decode the data (boolean).
        no_ssl: Don't try to establish a secure connection (to prevent SSLError).
        with_headers: Keep track of the response headers.
        config: Pass configuration values for output control.

    Returns:
        Response object or None in case of failed downloads and invalid results.

    """
    dl_function = _send_urllib_request if pycurl is None else _send_pycurl_request
    LOGGER.debug('sending request: %s', url)
    response = dl_function(url, no_ssl, with_headers, config)  # Response
    if not response:  # None or ""
        LOGGER.debug('request failed: %s', url)
        return None
    response.decode_data(decode)
    return response


def _pycurl_is_live_page(url):
    "Send a basic HTTP HEAD request with pycurl."
    # Initialize pycurl object
    curl = pycurl.Curl()
    # Set the URL and HTTP method (HEAD)
    curl.setopt(pycurl.URL, url.encode('utf-8'))
    curl.setopt(pycurl.CONNECTTIMEOUT, 10)
    # no SSL verification
    curl.setopt(pycurl.SSL_VERIFYPEER, 0)
    curl.setopt(pycurl.SSL_VERIFYHOST, 0)
    # Set option to avoid getting the response body
    curl.setopt(curl.NOBODY, True)
    # Perform the request
    try:
        curl.perform()
    except pycurl.error as err:
        LOGGER.debug('pycurl HEAD error: %s %s', url, err)
        return False
    # Get the response code
    page_exists = curl.getinfo(curl.RESPONSE_CODE) < 400
    # Clean up
    curl.close()
    return page_exists


def _urllib3_is_live_page(url):
    "Use courlan redirection test (based on urllib3) to send a HEAD request."
    try:
        _ = redirection_test(url)
    except Exception as err:
        LOGGER.debug('urllib3 HEAD error: %s %s', url, err)
        return False
    return True


def is_live_page(url):
    "Send a HTTP HEAD request without taking anything else into account."
    if pycurl is not None:
        return _pycurl_is_live_page(url) or _urllib3_is_live_page(url)
    return _urllib3_is_live_page(url)


def add_to_compressed_dict(inputlist, blacklist=None, url_filter=None, url_store=None, compression=False, verbose=False):
    '''Filter, convert input URLs and add them to domain-aware processing dictionary'''
    if url_store is None:
        url_store = UrlStore(
                        compressed=compression,
                        strict=False,
                        verbose=verbose
                    )

    inputlist = list(dict.fromkeys(inputlist))

    if blacklist:
        inputlist = [u for u in inputlist if URL_BLACKLIST_REGEX.sub('', u) not in blacklist]

    if url_filter:
        inputlist = [u for u in inputlist if any(f in u for f in url_filter)]

    url_store.add_urls(inputlist)
    return url_store


def load_download_buffer(url_store, sleep_time=5):
    '''Determine threading strategy and draw URLs respecting domain-based back-off rules.'''
    bufferlist = []
    while not bufferlist:
        bufferlist = url_store.get_download_urls(time_limit=sleep_time, max_urls=10**5)
        # add emptiness test or sleep?
        if not bufferlist:
            if url_store.done is True:
                break
            sleep(sleep_time)
    return bufferlist, url_store


def buffered_downloads(bufferlist, download_threads, decode=True, options=None):
    '''Download queue consumer, single- or multi-threaded.'''
    worker = partial(fetch_url, options=options) if decode else fetch_response
    with ThreadPoolExecutor(max_workers=download_threads) as executor:
        for chunk in make_chunks(bufferlist, 10000):
            future_to_url = {executor.submit(worker, url): url for url in chunk}
            for future in as_completed(future_to_url):
                # url and download result
                yield future_to_url[future], future.result()


def _send_pycurl_request(url, no_ssl, with_headers, config):
    '''Experimental function using libcurl and pycurl to speed up downloads'''
    # https://github.com/pycurl/pycurl/blob/master/examples/retriever-multi.py

    # init
    headerbytes = BytesIO()
    headers = _determine_headers(config)
    headerlist = ['Accept-Encoding: gzip, deflate', 'Accept: */*']
    for header, content in headers.items():
        headerlist.append(header + ': ' + content)

    # prepare curl request
    # https://curl.haxx.se/libcurl/c/curl_easy_setopt.html
    curl = pycurl.Curl()
    curl.setopt(pycurl.URL, url.encode('utf-8'))
    # share data
    curl.setopt(pycurl.SHARE, CURL_SHARE)
    curl.setopt(pycurl.HTTPHEADER, headerlist)
    # curl.setopt(pycurl.USERAGENT, '')
    curl.setopt(pycurl.FOLLOWLOCATION, 1)
    curl.setopt(pycurl.MAXREDIRS, config.getint('DEFAULT', 'MAX_REDIRECTS'))
    curl.setopt(pycurl.CONNECTTIMEOUT, config.getint('DEFAULT', 'DOWNLOAD_TIMEOUT'))
    curl.setopt(pycurl.TIMEOUT, config.getint('DEFAULT', 'DOWNLOAD_TIMEOUT'))
    curl.setopt(pycurl.MAXFILESIZE, config.getint('DEFAULT', 'MAX_FILE_SIZE'))
    curl.setopt(pycurl.NOSIGNAL, 1)

    if no_ssl is True:
        curl.setopt(pycurl.SSL_VERIFYPEER, 0)
        curl.setopt(pycurl.SSL_VERIFYHOST, 0)
    else:
        curl.setopt(pycurl.CAINFO, certifi.where())

    if with_headers:
        curl.setopt(pycurl.HEADERFUNCTION, headerbytes.write)

    # TCP_FASTOPEN
    # curl.setopt(pycurl.FAILONERROR, 1)
    # curl.setopt(pycurl.ACCEPT_ENCODING, '')

    # send request
    try:
        bufferbytes = curl.perform_rb()
    except pycurl.error as err:
        LOGGER.error('pycurl error: %s %s', url, err)
        # retry in case of SSL-related error
        # see https://curl.se/libcurl/c/libcurl-errors.html
        # errmsg = curl.errstr_raw()
        # additional error codes: 80, 90, 96, 98
        if no_ssl is False and err.args[0] in (35, 54, 58, 59, 60, 64, 66, 77, 82, 83, 91):
            LOGGER.debug('retrying after SSL error: %s %s', url, err)
            return _send_pycurl_request(url, True, with_headers, config)
        # traceback.print_exc(file=sys.stderr)
        # sys.stderr.flush()
        return None

    # additional info
    # ip_info = curl.getinfo(curl.PRIMARY_IP)

    resp = Response(bufferbytes, curl.getinfo(curl.RESPONSE_CODE), curl.getinfo(curl.EFFECTIVE_URL))
    curl.close()

    if with_headers:
        respheaders = {}
        # https://github.com/pycurl/pycurl/blob/master/examples/quickstart/response_headers.py
        for line in headerbytes.getvalue().decode("iso-8859-1", errors="replace").splitlines():
            # re.split(r'\r?\n') ?
            # This will botch headers that are split on multiple lines...
            if ':' not in line:
                continue
            # Break the header line into header name and value.
            name, value = line.split(':', 1)
            # Now we can actually record the header name and value.
            respheaders[name.strip()] = value.strip() # name.strip().lower() ?
        resp.store_headers(respheaders)

    return resp



"""
Module bundling all functions needed to scrape metadata from webpages.
"""

import json
import logging
import re

from copy import deepcopy

from courlan import extract_domain, get_base_url, is_valid_url, normalize_url, validate_url
from htmldate import find_date
from lxml.html import tostring

from .htmlprocessing import prune_unwanted_nodes
from .json_metadata import (extract_json, extract_json_parse_error,
                            normalize_json)
from .settings import set_date_params
from .utils import (line_processing, load_html, normalize_authors,
                    normalize_tags, trim, unescape)
from .xpaths import (AUTHOR_DISCARD_XPATHS, AUTHOR_XPATHS,
                     CATEGORIES_XPATHS, TAGS_XPATHS, TITLE_XPATHS)

LOGGER = logging.getLogger(__name__)
logging.getLogger('htmldate').setLevel(logging.WARNING)


class Document:
    "Defines a class to store all necessary data and metadata fields for extracted information."
    __slots__ = [
    'title', 'author', 'url', 'hostname', 'description', 'sitename',
    'date', 'categories', 'tags', 'fingerprint', 'id', 'license',
    'body', 'comments', 'commentsbody', 'raw_text', 'text',
    'language', 'image', 'pagetype', 'filedate'  # 'locale'?
    ]
    # consider dataclasses for Python 3.7+
    def __init__(self):
        for slot in self.__slots__:
            setattr(self, slot, None)

    def set_attributes(self, title, author, url, description, site_name, image, pagetype, tags):
        "Helper function to (re-)set a series of attributes."
        if title:
            self.title = title
        if author:
            self.author = author
        if url:
            self.url = url
        if description:
            self.description = description
        if site_name:
            self.sitename = site_name
        if image:
            self.image = image
        if pagetype:
            self.pagetype = pagetype
        if tags:
            self.tags = tags

    def clean_and_trim(self):
        "Limit text length and trim the attributes."
        for slot in self.__slots__:
            value = getattr(self, slot)
            if isinstance(value, str):
                # length
                if len(value) > 10000:
                    new_value = value[:9999] + '…'
                    setattr(self, slot, new_value)
                    value = new_value
                # HTML entities, remove spaces and control characters
                value = line_processing(unescape(value))
                setattr(self, slot, value)

    def as_dict(self):
        "Convert the document to a dictionary."
        return {
            attr: getattr(self, attr, None)
            for attr in self.__slots__
        }


JSON_MINIFY = re.compile(r'("(?:\\"|[^"])*")|\s')

HTMLTITLE_REGEX = re.compile(r'^(.+)?\s+[–•·—|⁄*⋆~‹«<›»>:-]\s+(.+)$')  # part without dots?
HTML_STRIP_TAG = re.compile(r'(<!--.*?-->|<[^>]*>)')

LICENSE_REGEX = re.compile(r'/(by-nc-nd|by-nc-sa|by-nc|by-nd|by-sa|by|zero)/([1-9]\.[0-9])')
TEXT_LICENSE_REGEX = re.compile(r'(cc|creative commons) (by-nc-nd|by-nc-sa|by-nc|by-nd|by-sa|by|zero) ?([1-9]\.[0-9])?', re.I)

METANAME_AUTHOR = {
    'article:author', 'atc-metaauthor', 'author', 'authors', 'byl', 'citation_author',
    'creator', 'dc.creator', 'dc.creator.aut', 'dc:creator',
    'dcterms.creator', 'dcterms.creator.aut', 'dcsext.author', 'parsely-author',
    'rbauthors', 'sailthru.author', 'shareaholic:article_author_name'
}  # questionable: twitter:creator
METANAME_DESCRIPTION = {
    'dc.description', 'dc:description',
    'dcterms.abstract', 'dcterms.description',
    'description', 'sailthru.description', 'twitter:description'
}
METANAME_PUBLISHER = {
    'article:publisher', 'citation_journal_title', 'copyright',
    'dc.publisher', 'dc:publisher', 'dcterms.publisher',
    'publisher', 'sailthru.publisher', 'rbpubname', 'twitter:site'
}  # questionable: citation_publisher
METANAME_TAG = {
    'citation_keywords', 'dcterms.subject', 'keywords', 'parsely-tags',
    'shareaholic:keywords', 'tags'
}
METANAME_TITLE = {
    'citation_title', 'dc.title', 'dcterms.title', 'fb_title',
    'headline', 'parsely-title', 'sailthru.title', 'shareaholic:title',
    'rbtitle', 'title', 'twitter:title'
}
METANAME_URL = {
    'rbmainurl', 'twitter:url'
}
METANAME_IMAGE = {
    'image', 'og:image', 'og:image:url', 'og:image:secure_url',
    'twitter:image', 'twitter:image:src'
}
OG_AUTHOR = {'og:author', 'og:article:author'}
PROPERTY_AUTHOR = {'author', 'article:author'}
TWITTER_ATTRS = {'twitter:site', 'application-name'}

# also interesting: article:section

EXTRA_META = {'charset', 'http-equiv', 'property'}


def check_authors(authors, author_blacklist):
    "Check if the authors string correspond to expected values."
    author_blacklist = {a.lower() for a in author_blacklist}
    new_authors = [
        author.strip()
        for author in authors.split(';')
        if author.strip().lower() not in author_blacklist
    ]
    if new_authors:
        return '; '.join(new_authors).strip('; ')
    return None


def extract_meta_json(tree, metadata):
    '''Parse and extract metadata from JSON-LD data'''
    for elem in tree.xpath('.//script[@type="application/ld+json" or @type="application/settings+json"]'):
        if not elem.text:
            continue
        element_text = normalize_json(JSON_MINIFY.sub(r'\1', elem.text))
        try:
            schema = json.loads(element_text)
            metadata = extract_json(schema, metadata)
        except json.JSONDecodeError:
            metadata = extract_json_parse_error(element_text, metadata)
    return metadata


def extract_opengraph(tree):
    '''Search meta tags following the OpenGraph guidelines (https://ogp.me/)'''
    title, author, url, description, site_name, image, pagetype = (None,) * 7
    # detect OpenGraph schema
    for elem in tree.xpath('.//head/meta[starts-with(@property, "og:")]'):
        # safeguard
        if not elem.get('content') or elem.get('content').isspace():
            continue
        # site name
        if elem.get('property') == 'og:site_name':
            site_name = elem.get('content')
        # blog title
        elif elem.get('property') == 'og:title':
            title = elem.get('content')
        # orig URL
        elif elem.get('property') == 'og:url':
            if is_valid_url(elem.get('content')):
                url = elem.get('content')
        # description
        elif elem.get('property') == 'og:description':
            description = elem.get('content')
        # og:author
        elif elem.get('property') in OG_AUTHOR:
            author = normalize_authors(None, elem.get('content'))
        # image default
        elif elem.get('property') == 'og:image':
            image = elem.get('content')
        # image url
        elif elem.get('property') == 'og:image:url':
            image = elem.get('content')
        # image secure url
        elif elem.get('property') == 'og:image:secure_url':
            image = elem.get('content')
        # og:type
        elif elem.get('property') == 'og:type':
            pagetype = elem.get('content')
        # og:locale
        # elif elem.get('property') == 'og:locale':
        #    pagelocale = elem.get('content')
    return title, author, url, description, site_name, image, pagetype


def examine_meta(tree):
    '''Search meta tags for relevant information'''
    metadata = Document()  # alt: Metadata()
    # bootstrap from potential OpenGraph tags
    title, author, url, description, site_name, image, pagetype = extract_opengraph(tree)
    # test if all values not assigned in the following have already been assigned
    if all((title, author, url, description, site_name, image)):
        metadata.set_attributes(title, author, url, description, site_name, image, pagetype, None)  # tags
        return metadata
    tags, backup_sitename = [], None
    # skim through meta tags
    for elem in tree.iterfind('.//head/meta[@content]'):
        # content
        if not elem.get('content') or elem.get('content').isspace():
            continue
        content_attr = HTML_STRIP_TAG.sub('', elem.get('content'))
        # image info
        # ...
        # property
        if 'property' in elem.attrib:
            # no opengraph a second time
            if elem.get('property').startswith('og:'):
                continue
            if elem.get('property') == 'article:tag':
                tags.append(normalize_tags(content_attr))
            elif elem.get('property') in PROPERTY_AUTHOR:
                author = normalize_authors(author, content_attr)
            elif elem.get('property') == 'article:publisher':
                site_name = site_name or content_attr
            elif elem.get('property') in METANAME_IMAGE:
                image = image or content_attr
        # name attribute
        elif 'name' in elem.attrib:
            name_attr = elem.get('name').lower()
            # author
            if name_attr in METANAME_AUTHOR:
                author = normalize_authors(author, content_attr)
            # title
            elif name_attr in METANAME_TITLE:
                title = title or content_attr
            # description
            elif name_attr in METANAME_DESCRIPTION:
                description = description or content_attr
            # site name
            elif name_attr in METANAME_PUBLISHER:
                site_name = site_name or content_attr
            # twitter
            elif name_attr in TWITTER_ATTRS or 'twitter:app:name' in elem.get('name'):
                backup_sitename = content_attr
            # url
            elif name_attr == 'twitter:url':
                if url is None and is_valid_url(content_attr):
                    url = content_attr
            # keywords
            elif name_attr in METANAME_TAG:  # 'page-topic'
                tags.append(normalize_tags(content_attr))
        elif 'itemprop' in elem.attrib:
            if elem.get('itemprop') == 'author':
                author = normalize_authors(author, content_attr)
            elif elem.get('itemprop') == 'description':
                description = description or content_attr
            elif elem.get('itemprop') == 'headline':
                title = title or content_attr
            # to verify:
            # elif elem.get('itemprop') == 'name':
            #    if title is None:
            #        title = elem.get('content')
        # other types
        elif all(
            key not in elem.attrib
            for key in EXTRA_META
        ):
            LOGGER.debug('unknown attribute: %s',
                         tostring(elem, pretty_print=False, encoding='unicode').strip())
    # backups
    if site_name is None and backup_sitename is not None:
        site_name = backup_sitename
    # copy
    metadata.set_attributes(title, author, url, description, site_name, image, pagetype, tags)
    return metadata


def extract_metainfo(tree, expressions, len_limit=200):
    '''Extract meta information'''
    # try all XPath expressions
    for expression in expressions:
        # examine all results
        i = 0
        for elem in expression(tree):
            content = trim(' '.join(elem.itertext()))
            if content and 2 < len(content) < len_limit:
                return content
            i += 1
        if i > 1:
            LOGGER.debug('more than one invalid result: %s %s', expression, i)
    return None


def examine_title_element(tree):
    '''Extract text segments out of main <title> element.'''
    title, first, second = None, None, None
    try:
        title = trim(tree.xpath('.//head//title')[0].text_content())
        mymatch = HTMLTITLE_REGEX.match(title)
        if mymatch is not None:
            first = mymatch[1] or None
            second = mymatch[2] or None
    except IndexError:
        LOGGER.debug('no main title found')
    return title, first, second


def extract_title(tree):
    '''Extract the document title'''
    # only one h1-element: take it
    h1_results = tree.findall('.//h1')
    if len(h1_results) == 1:
        title = trim(h1_results[0].text_content())
        if len(title) > 0:
            return title
    # extract using x-paths
    title = extract_metainfo(tree, TITLE_XPATHS)
    if title is not None:
        return title
    # extract using title tag
    title, first, second = examine_title_element(tree)
    if first is not None and '.' not in first:
        return first
    if second is not None and '.' not in second:
        return second
    # take first h1-title
    if h1_results:
        return h1_results[0].text_content()
    # take first h2-title
    try:
        title = tree.xpath('.//h2')[0].text_content()
    except IndexError:
        LOGGER.debug('no h2 title found')
    return title


def extract_author(tree):
    '''Extract the document author(s)'''
    subtree = prune_unwanted_nodes(deepcopy(tree), AUTHOR_DISCARD_XPATHS)
    author = extract_metainfo(subtree, AUTHOR_XPATHS, len_limit=120)
    if author:
        author = normalize_authors(None, author)
    # copyright?
    return author


def extract_url(tree, default_url=None):
    '''Extract the URL from the canonical link'''
    url = None
    # https://www.tutorialrepublic.com/html-reference/html-base-tag.php
    # try canonical link first
    element = tree.find('.//head//link[@rel="canonical"][@href]')
    if element is not None:
        url = element.attrib['href']
    # try default language link
    else:
        for element in tree.iterfind('.//head//link[@rel="alternate"][@hreflang]'):
            if element.attrib['hreflang'] == 'x-default':
                LOGGER.debug(tostring(element, pretty_print=False, encoding='unicode').strip())
                url = element.attrib['href']
    # add domain name if it's missing
    if url is not None and url.startswith('/'):
        for element in tree.iterfind('.//head//meta[@content]'):
            if 'name' in element.attrib:
                attrtype = element.attrib['name']
            elif 'property' in element.attrib:
                attrtype = element.attrib['property']
            else:
                continue
            if attrtype.startswith('og:') or attrtype.startswith('twitter:'):
                base_url = get_base_url(element.attrib['content'])
                if base_url:
                    # prepend URL
                    url = base_url + url
                    break
    # sanity check: don't return invalid URLs
    if url is not None:
        validation_result, parsed_url = validate_url(url)
        url = None if validation_result is False else normalize_url(parsed_url)
    return url or default_url


def extract_sitename(tree):
    '''Extract the name of a site from the main title (if it exists)'''
    _, first, second = examine_title_element(tree)
    if first is not None and '.' in first:
        return first
    if second is not None and '.' in second:
        return second
    return None


def extract_catstags(metatype, tree):
    '''Find category and tag information'''
    results = []
    regexpr = '/' + metatype + '[s|ies]?/'
    xpath_expression = CATEGORIES_XPATHS if metatype == 'category' else TAGS_XPATHS
    # search using custom expressions
    for catexpr in xpath_expression:
        results.extend(
            elem.text_content()
            for elem in catexpr(tree)
            if re.search(regexpr, elem.attrib['href'])
        )
        if results:
            break
    # category fallback
    if metatype == 'category' and not results:
        for element in tree.xpath('.//head//meta[@property="article:section" or contains(@name, "subject")][@content]'):
            results.append(element.attrib['content'])
        # optional: search through links
        #if not results:
        #    for elem in tree.xpath('.//a[@href]'):
        #        search for 'category'
    return [r for r in dict.fromkeys(line_processing(x) for x in results if x) if r]


def parse_license_element(element, strict=False):
    '''Probe a link for identifiable free license cues.
       Parse the href attribute first and then the link text.'''
   # look for Creative Commons elements
    match = LICENSE_REGEX.search(element.get('href'))
    if match:
        return 'CC ' + match[1].upper() + ' ' + match[2]
    if element.text is not None:
        # just return the anchor text without further ado
        if strict is False:
            return trim(element.text)
        # else: check if it could be a CC license
        match = TEXT_LICENSE_REGEX.search(element.text)
        if match:
            return match[0]
    return None


def extract_license(tree):
    '''Search the HTML code for license information and parse it.'''
    result = None
    # look for links labeled as license
    for element in tree.findall('.//a[@rel="license"][@href]'):
        result = parse_license_element(element, strict=False)
        if result is not None:
            break
    # probe footer elements for CC links
    if result is None:
        for element in tree.xpath(
            './/footer//a[@href]|.//div[contains(@class, "footer") or contains(@id, "footer")]//a[@href]'
        ):
            result = parse_license_element(element, strict=True)
            if result is not None:
                break
    return result


def extract_image(tree):
    '''Search meta tags following the OpenGraph guidelines (https://ogp.me/)
       and search meta tags with Twitter Image'''

    for elem in tree.xpath('.//head/meta[@property="og:image" or @property="og:image:url"][@content]'):
        return elem.get('content')

    for elem in tree.xpath('.//head/meta[@property="twitter:image" or @property="twitter:image:src"][@content]'):
        return elem.get('content')

    return None


def extract_metadata(filecontent, default_url=None, date_config=None, extensive=True, author_blacklist=None):
    """Main process for metadata extraction.

    Args:
        filecontent: HTML code as string.
        default_url: Previously known URL of the downloaded document.
        date_config: Provide extraction parameters to htmldate as dict().
        author_blacklist: Provide a blacklist of Author Names as set() to filter out authors.

    Returns:
        A trafilatura.metadata.Document containing the extracted metadata information or None.
        trafilatura.metadata.Document has .as_dict() method that will return a copy as a dict.
    """
    # init
    if author_blacklist is None:
        author_blacklist = set()
    if not date_config:
        date_config = set_date_params(extensive)
    # load contents
    tree = load_html(filecontent)
    if tree is None:
        return None
    # initialize dict and try to strip meta tags
    metadata = examine_meta(tree)
    # to check: remove it and replace with author_blacklist in test case
    if metadata.author is not None and ' ' not in metadata.author:
        metadata.author = None
    # fix: try json-ld metadata and override
    try:
        metadata = extract_meta_json(tree, metadata)
    # todo: fix bugs in json_metadata.py
    except TypeError as err:
        LOGGER.warning('error in JSON metadata extraction: %s', err)
    # title
    if metadata.title is None:
        metadata.title = extract_title(tree)
    # check author in blacklist
    if metadata.author is not None and len(author_blacklist) > 0:
        metadata.author = check_authors(metadata.author, author_blacklist)
    # author
    if metadata.author is None:
        metadata.author = extract_author(tree)
    # recheck author in blacklist
    if metadata.author is not None and len(author_blacklist) > 0:
        metadata.author = check_authors(metadata.author, author_blacklist)
    # url
    if metadata.url is None:
        metadata.url = extract_url(tree, default_url)
    # hostname
    if metadata.url is not None:
        metadata.hostname = extract_domain(metadata.url, fast=True)
    # image
    if metadata.image is None:
        metadata.image = extract_image(tree)
    # extract date with external module htmldate
    date_config['url'] = metadata.url
    metadata.date = find_date(tree, **date_config)
    # sitename
    if metadata.sitename is None:
        metadata.sitename = extract_sitename(tree)
    if metadata.sitename is not None:
        # fix: take 1st element (['Westdeutscher Rundfunk'])
        if isinstance(metadata.sitename, list):
            metadata.sitename = metadata.sitename[0]
        # hotfix: probably an error coming from json_metadata (#195)
        elif isinstance(metadata.sitename, dict):
            metadata.sitename = str(metadata.sitename)
        if metadata.sitename.startswith('@'):
            # scrap Twitter ID
            metadata.sitename = re.sub(r'^@', '', metadata.sitename)
        # capitalize
        try:
            if (
                '.' not in metadata.sitename
                and not metadata.sitename[0].isupper()
            ):
                metadata.sitename = metadata.sitename.title()
        # fix for empty name
        except IndexError as err:
            LOGGER.warning('error in sitename extraction: %s', err)
    # use URL
    elif metadata.url:
        mymatch = re.match(r'https?://(?:www\.|w[0-9]+\.)?([^/]+)', metadata.url)
        if mymatch:
            metadata.sitename = mymatch[1]
    # categories
    if not metadata.categories:
        metadata.categories = extract_catstags('category', tree)
    # tags
    if not metadata.tags:
        metadata.tags = extract_catstags('tag', tree)
    # license
    metadata.license = extract_license(tree)
    # safety checks
    metadata.filedate = date_config["max_date"]
    metadata.clean_and_trim()
    # return result
    return metadata
